#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 25 15:21:01 2019

@author: usingh
"""

from pyrpipe import pyrpipe_utils as pu
from pyrpipe import pyrpipe_engine as pe
import os

class Assembly:
    """This class represents an abstract parent class for all programs which can perfrom transcripts assembly.
    """
    def __init__(self):
        self.category="Assembler"
    def perform_assembly(bam_file):
        """Function to perform assembly using a bam file as input. Inherited by all children.
        
                
        :param bam_file: path to input BAM
        :type bam_file: string
        
        :return: path to output GTF or output directory depending on the specific assembly program.
        :rtype: string
        """
        pass

class Stringtie(Assembly):
    """This class represents Stringtie program for transcript assembly.
        
        Parameters
        ----------
        
        reference_gtf: string
            Path to the reference gtf file. If a valid gtf file is provided the option -G will be set to the gtf file. This can't
            be overriden later when calling functions of this class.
        kwargs: dict
            Options passed to stringtie. Some of these could be overridden later when calling functions of this class.
            Format to pass the arguments:
        """
    def __init__(self,reference_gtf="",**kwargs):
        
        super().__init__()
        self.program_name="stringtie"
        #check if stringtie exists
        if not pe.check_dependencies([self.program_name]):
            raise Exception("ERROR: "+ self.program_name+" not found.")
        self.valid_args_list=['-G','--version','--conservative','--rf','--fr','-o','-l',
                            '-f','-L','-m','-a','-j','-t','-c','-s','-v','-g','-M',
                            '-p','-A','-B','-b','-e','-x','-u','-h','--merge','-F','-T','-i']
        
        #keep the passed arguments
        self.passed_args_dict=kwargs
        
        #check the reference GTF
        if len(reference_gtf)>0 and pu.check_files_exist(reference_gtf):
            self.reference_gtf=reference_gtf
            self.passed_args_dict['-G']=reference_gtf
        
    def perform_assembly(self,bam_file,out_dir="",out_suffix="_stringtie",overwrite=True,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to run stringtie using a bam file.
                
        Parameters
        ----------
        
        bam_file: string
            path to the bam file
        out_suffix: string
            Suffix for the output gtf file
        overwrite: bool
            Overwrite if output file already exists.
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to stringtie. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).
            
        :return: Returns the path to output GTF file
        :rtype: string
        """
        
        #create path to output file
        fname=pu.get_file_basename(bam_file)
        
        if not out_dir:
            out_dir=pu.get_file_directory(bam_file)
            
        out_gtf_file=os.path.join(out_dir,fname+out_suffix+".gtf")
        
        """
        Handle overwrite
        """
        if not overwrite:
            #check if file exists. return if yes
            if os.path.isfile(out_gtf_file):
                print("The file "+out_gtf_file+" already exists. Exiting..")
                return out_gtf_file
        
        #Add output file name and input bam
        new_opts={"-o":out_gtf_file,"--":(bam_file,)}
        merged_opts={**kwargs,**new_opts}
        
        #call stringtie
        status=self.run_stringtie(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**merged_opts)
        
        if status:
            #check if sam file is present in the location directory of sraOb
            if pu.check_files_exist(out_gtf_file):
                return out_gtf_file
        else:
            return ""
        
    def stringtie_merge(self,*args,out_suffix="_stringtieMerge",overwrite=True,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to run stringtie merge.
        
        Parameters
        ----------
        
        args: tuple
            path to gtf files to merge
        out_suffix: string
            Suffix for output gtf file name
        overwrite: bool
            Overwrite if output file already exists.
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to stringtie. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).

        :return: Returns the path to the merged GTF file
        :rtype: string
        """
        
        if len(args) < 1:
            print("ERROR: No input gtf for stringtie merge.")
            return ""
        
        #create path to output sam file
        fname=pu.get_file_basename(args[0])
        out_dir=pu.get_file_directory(args[0])
        out_gtf_file=os.path.join(out_dir,fname+out_suffix+".gtf")
        
        if not overwrite:
            #check if file exists. return if yes
            if os.path.isfile(out_gtf_file):
                print("The file "+out_gtf_file+" already exists. Exiting..")
                return out_gtf_file
        
        #Add merge flag, output file name and input bam
        new_opts={"--merge":"","-o":out_gtf_file,"--":args}
        
        merged_opts={**kwargs,**new_opts}
        
        #call stringtie
        status=self.run_stringtie(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**merged_opts)
        
        if status:
            #check if sam file is present in the location directory of sraOb
            if pu.check_files_exist(out_gtf_file):
                return out_gtf_file
        else:
            return ""
        
        
        
            
    
    def run_stringtie(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running stringtie. This can be used to run stringtie without using perform_assembly() function.
        
        Parameters
        ----------
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to stringtie. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).

        :return: Returns the status of stringtie command.
        :rtype: bool
        """
            
        #override existing arguments
        merged_args_dict={**self.passed_args_dict,**kwargs}
       
        stie_cmd=['stringtie']
        #add options
        stie_cmd.extend(pu.parse_unix_args(self.valid_args_list,merged_args_dict))        
        
                
        #start ececution
        status=pe.execute_command(stie_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("stringtie failed")
        
        #return status
        return status
    
    
    
class Cufflinks(Assembly):
    """This class represents cufflinks
    
    kwargs: dict
        Parameters passed to cufflinks
    """
    def __init__(self,reference_gtf="",**kwargs):
        
        super().__init__()
        self.program_name="cufflinks"
        #check if stringtie exists
        if not pe.check_dependencies([self.program_name]):
            raise Exception("ERROR: "+ self.program_name+" not found.")
            
        
        #define valid arguments
        self.cufflinksArgsList=['-h','--help','-o','--output-dir','-p','--num-threads','--seed','-G','--GTF','-g','--GTF-guide','-M','--mask-file','-b','--frag-bias-correct','-u','--multi-read-correct','--library-type','--library-norm-method',
'-m','--frag-len-mean','-s','--frag-len-std-dev','--max-mle-iterations','--compatible-hits-norm','--total-hits-norm','--num-frag-count-draws','--num-frag-assign-draws','--max-frag-multihits','--no-effective-length-correction',
'--no-length-correction','-N','--upper-quartile-norm','--raw-mapped-norm','-L','--label','-F','--min-isoform-fraction','-j','--pre-mrna-fraction','-I','--max-intron-length','-a','--junc-alpha','-A','--small-anchor-fraction',
'--min-frags-per-transfrag','--overhang-tolerance','--max-bundle-length','--max-bundle-frags','--min-intron-length','--trim-3-avgcov-thresh','--trim-3-dropoff-frac','--max-multiread-fraction','--overlap-radius',
'--no-faux-reads','--3-overhang-tolerance','--intron-overhang-tolerance','-v','--verbose','-q','--quiet','--no-update-check']

        self.cuffcompareArgsList=['-h','-i','-r','-R','-Q','-M','-N','-s','-e','-d','-p','-C','-F','-G','-T','-V']
        self.cuffquantArgsList=['-o','--output-dir','-p','--num-threads','-M','--mask-file','-b','--frag-bias-correct','-u','--multi-read-correct','--library-type','-m','--frag-len-mean','-s','--frag-len-std-dev','-c','--min-alignment-count',
'--max-mle-iterations','-v','--verbose','-q','--quiet','--seed','--no-update-check','--max-bundle-frags','--max-frag-multihits','--no-effective-length-correction','--no-length-correction','--read-skip-fraction',
'--no-read-pairs','--trim-read-length','--no-scv-correction']
        self.cuffdiffArgsList=['-o','--output-dir','-L','--labels','--FDR','-M','--mask-file','-C','--contrast-file','-b','--frag-bias-correct','-u','--multi-read-correct','-p','--num-threads','--no-diff','--no-js-tests','-T','--time-series',
'--library-type','--dispersion-method','--library-norm-method','-m','--frag-len-mean','-s','--frag-len-std-dev','-c','--min-alignment-count','--max-mle-iterations','--compatible-hits-norm','--total-hits-norm',
' -v','--verbose','-q','--quiet','--seed','--no-update-check','--emit-count-tables','--max-bundle-frags','--num-frag-count-draws','--num-frag-assign-draws','--max-frag-multihits','--min-outlier-p','--min-reps-for-js-test',
'--no-effective-length-correction','--no-length-correction','-N','--upper-quartile-norm','--geometric-norm','--raw-mapped-norm','--poisson-dispersion','--read-skip-fraction','--no-read-pairs','--trim-read-length','--no-scv-correction']
        self.cuffnormArgsList=['-o','--output-dir','-L','--labels','--norm-standards-file','-p','--num-threads','--library-type','--library-norm-method','--output-format','--compatible-hits-norm','--total-hits-norm','-v','--verbose','-q','--quiet','--seed','--no-update-check']
        self.cuffmergeArgsList=['h','--help','-o','-g','–-ref-gtf','-p','–-num-threads','-s','-–ref-sequence']
        
        self.valid_args_list=pu.get_union(self.cufflinksArgsList,self.cuffcompareArgsList,self.cuffquantArgsList,self.cuffdiffArgsList,self.cuffnormArgsList,self.cuffmergeArgsList)
        
        #keep the passed arguments
        self.passed_args_dict=kwargs
        
        #check the reference GTF
        if len(reference_gtf)>0 and pu.check_files_exist(reference_gtf):
            self.reference_gtf=reference_gtf
            self.passed_args_dict['-g']=reference_gtf
    
    
    def perform_assembly(self,bam_file,out_dir="",out_suffix="_cufflinks",overwrite=True,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to run cufflinks with BAM file as input.
                
        Parameters
        ----------
        bam_file: string
            path to bam file
        out_dir: 
            output directory
        out_suffix: string
            Suffix for the output gtf file
        overwrite: bool
            Overwrite if output file already exists.
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession.
        kwargs: dict
            Options to pass to cufflinks. This will override the existing options self.passed_args_dict (only replace existing arguments and not replace all the arguments).

        :return: Returns the path to output GTF file
        :rtype: string       
        """
        
        #create path to output file
        fname=pu.get_file_basename(bam_file)
        if not out_dir:
            out_dir=pu.get_file_directory(bam_file)
        else:
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
        out_gtf_file=os.path.join(out_dir,fname+out_suffix+".gtf")
        
        """
        Handle overwrite
        """
        if not overwrite:
            #check if file exists. return if yes
            if os.path.isfile(out_gtf_file):
                print("The file "+out_gtf_file+" already exists. Exiting..")
                return out_gtf_file
            
        #Add output file name and input bam
        new_opts={"-o":out_dir,"--":(bam_file,)}
        merged_opts={**kwargs,**new_opts}
        
        #call cufflinks
        status=self.run_cufflinks(verbose,quiet,logs,objectid,**merged_opts)
        
        if status:
            #move out_dir/transcripts.gtf to outfile
            pe.move_file(os.path.join(out_dir,"transcripts.gtf"),out_gtf_file)
            #check if sam file is present in the location directory of sraOb
            if pu.check_files_exist(out_gtf_file):
                return out_gtf_file
        else:
            return ""
    
    def run_cuff(self,command,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running cuff* commands
        
        Parameters
        ----------
        
        command: string
            the command name
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession.
        kwargs: dict
            Options passed to command
        
        :return: Returns the status of the command.
        :rtype: bool
        """
        validCommands=['cuffcompare','cuffdiff', 'cufflinks', 'cuffmerge', 'cuffnorm', 'cuffquant']
        if command in validCommands:
            #override existing arguments
            merged_args_dict={**self.passed_args_dict,**kwargs}
       
            cuff_cmd=[command]
            #add options
            cuff_cmd.extend(pu.parse_unix_args(self.valid_args_list,merged_args_dict))        
                  
            #start ececution
            status=pe.execute_command(cuff_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
            if not status:
                pu.print_boldred("cufflinks failed")
                #return status
            return status
        else:
            pu.print_boldred("Unknown command {}"+command)
            return False
    
    
    def run_cufflinks(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running cufflinks
        
        Parameters
        ----------
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession.
        kwargs: dict
            Options passed to cufflinks
        
        :return: Returns the status of cufflinks command.
        :rtype: bool
        """
            
        #override existing arguments
        merged_args_dict={**self.passed_args_dict,**kwargs}
       
        cufflinks_cmd=['cufflinks']
        #add options
        cufflinks_cmd.extend(pu.parse_unix_args(self.valid_args_list,merged_args_dict))        
        
        
        #start ececution
        status=pe.execute_command(cufflinks_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("cufflinks failed")
        #return status
        return status
    
    
    
class Trinity(Assembly):
    """This class represents Trinity RNA-Seq assembler
    
    kwargs: dict
        parametrs passed to Trinity
    """
    def __init__(self,**kwargs):
        
        super().__init__()
        self.program_name="Trinity"
        self.dep_list=[self.program_name,'jellyfish','bowtie2']
        #check if trinity exists
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.program_name+" not found.")
        
        
        self.valid_args_list=['--seqType','--max_memory','--left','--right','--single','--SS_lib_type','--CPU','--min_contig_length',
                              '--long_reads','--genome_guided_bam','--jaccard_clip','--trimmomatic','--normalize_reads','--no_distributed_trinity_exec',
                              '--output','--full_cleanup','--cite','--verbose','--version','--show_full_usage_info','--KMER_SIZE','--prep','--no_cleanup',
                              '--no_version_check','--min_kmer_cov','--inchworm_cpu','--no_run_inchworm','--max_reads_per_graph','--min_glue','--no_bowtie',
                              '--no_run_chrysalis','--bfly_opts','--PasaFly','--CuffFly','--group_pairs_distance','--path_reinforcement_distance','--no_path_merging',
                              '--min_per_id_same_path','--max_diffs_same_path','--max_internal_gap_same_path','--bflyHeapSpaceMax','--bflyHeapSpaceInit',
                              '--bflyGCThreads','--bflyCPU','--bflyCalculateCPU','--bfly_jar','--quality_trimming_params','--normalize_max_read_cov',
                              '--normalize_by_read_set','--genome_guided_max_intron','--genome_guided_min_coverage','--genome_guided_min_reads_per_partition',
                              '--grid_conf','--grid_node_CPU','--grid_node_max_memory']

        
        #keep the passed arguments
        self.passed_args_dict=kwargs
    
    
    def perform_assembly(self,sra_object=None,bam_file=None,out_dir="trinity_out_dir",max_memory="2G",max_intron=10000,overwrite=True,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to run trinity with sra object or BAM file as input.
                
        Parameters
        ----------
        
        sra_object: SRA
            object of SRA class
        bam_file: string
            path to bam file
        out_dir: string
            path to out directory
        max_memory: string
            Max memory argument e.g. "2G"
        max_intron: int
            specify the "--genome_guided_max_intron" argument
        overwrite: bool
            Overwrite if output file already exists
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession.
        
        kwargs: dict
            Options to pass to Trinity. This will override the existing options self.passed_args_dict (only replace existing arguments and not replace all the arguments).

        :return: Return the path to output GTF file
        :rtype: string
        """
        
        #add trinity to outdir
        if "trinity" not in out_dir:
            out_dir+="_trinity"
        
        new_opts={}
        if sra_object is not None:
            parent_dir=sra_object.location
            out_dir=os.path.join(parent_dir,out_dir)
            if sra_object.layout == 'PAIRED':
                new_opts={"--seqType":"fq","--left":sra_object.localfastq1Path,"--right":sra_object.localfastq2Path,"--output":out_dir,"--max_memory":max_memory}
            else:
                new_opts={"--seqType":"fq","--single":sra_object.localfastqPath,"--output":out_dir,"--max_memory":max_memory}
        elif bam_file is not None:
            if not pu.check_files_exist(bam_file):
                pu.print_boldred("Input to trinity does not exist:"+bam_file)
                return ""
            parent_dir=pu.get_file_directory(bam_file)
            out_dir=os.path.join(parent_dir,out_dir)
            new_opts={"--genome_guided_bam":bam_file,"--output":out_dir,"--max_memory":max_memory,"--genome_guided_max_intron":max_intron}
        else:
            pu.print_boldred("Please provide valid input to run trinity")
            return ""
        
        merged_opts={**kwargs,**new_opts}
        
        #call trinity
        status=self.run_trinity(verbose,quiet,logs,objectid,**merged_opts)
        
        if status:
            #check out dir
            if pu.check_paths_exist(out_dir):
                return out_dir
        else:
            return ""

    
    
    def run_trinity(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running trinity
        
        Parameters
        ----------
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession.
        kwargs: dict
            Options passed to trinity

        :return: Return the status of trinity command.
        :rtype: bool
        """
            
        #override existing arguments
        merged_args_dict={**self.passed_args_dict,**kwargs}
       
        trinity_cmd=['Trinity']
        #add options
        trinity_cmd.extend(pu.parse_unix_args(self.valid_args_list,merged_args_dict))        
        
        
        #start ececution
        status=pe.execute_command(trinity_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("trinity failed")
        #return status
        return status
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 25 17:48:00 2019

@author: usingh
"""

from pyrpipe import pyrpipe_utils as pu
from pyrpipe import pyrpipe_engine as pe
import os

class RNASeqQC:
    """This is an abstract parent class for fastq quality control programs.
    """
    def __init__(self):
        self.category="RNASeqQC"
        
    def perform_qc(self):
        pass

class Trimgalore(RNASeqQC):
    """This class represents trimgalore
    
    Parameters
    ----------
        
        kwargs:
            trim_galore arguments.
    """
    def __init__(self,**kwargs):
        """
        
        """
        
        #run super to inherit parent class properties
        super().__init__() 
        self.programName="trim_galore"
        self.dep_list=[self.programName,'cutadapt']
        self.valid_args=['--cores','-v','-q','--phred33','--phred64','--fastqc','--fastqc_args','-a','-a2',
                            '--illumina','--nextera','--small_rna','--consider_already_trimmed',
                            '--max_length','--stringency','-e','--gzip','--dont_gzip','--length',
                            '--max_n','--trim-n','-o','--no_report_file','--suppress_warn',
                            '--clip_R1','--clip_R2','--three_prime_clip_R1','--three_prime_clip_R2',
                            '--2colour','--path_to_cutadapt','--basename','-j','--hardtrim5','--hardtrim3',
                            '--clock','--polyA','--rrbs','--non_directional','--keep','--paired','-t',
                            '--retain_unpaired','-r1','-r2']
        #check if hisat2 exists
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
            
        #initialize the passed arguments
        self.passedArgumentDict=kwargs
        
        
            
    def perform_qc(self,sra_object,out_dir="",out_suffix="_trimgalore",verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to perform qc using trimgalore.
        The function perform_qc() is consistent for all QC classess.
        
        Parameters
        ----------
        
        sra_object: SRA
            An SRA object whose fastq files will be used
        out_suffix: string
            Suffix for the output sam file
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to trimgalore. This will override the existing options 

            :return: Returns the path of fastq files after QC. tuple has one item for single end files and two for paired.
            :rtype: tuple
        """
        
        if not out_dir:
            out_dir=sra_object.location
        else:
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
        
        #create new options based on parametrs
        newOpts={}
        #get layout
        if sra_object.layout=='PAIRED':
            fq1=sra_object.localfastq1Path
            fq2=sra_object.localfastq2Path
            out_file1=os.path.join(out_dir,pu.get_file_basename(fq1)+out_suffix+".fastq")
            out_file2=os.path.join(out_dir,pu.get_file_basename(fq2)+out_suffix+".fastq")
            newOpts={"--paired":"","--":(fq1,fq2),"-o":out_dir}
            mergedOpts={**kwargs,**newOpts}
            #run trimgalore
            self.run_trimgalore(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
            """
            running trim galore will create two files named <input>_val_1.fq and <input>_val_2.fq
            move these files to the specified out files
            """
            oldFile1=os.path.join(out_dir,pu.get_file_basename(fq1)+"_val_1.fq")
            oldFile2=os.path.join(out_dir,pu.get_file_basename(fq2)+"_val_2.fq")
            
            pe.move_file(oldFile1,out_file1)
            pe.move_file(oldFile2,out_file2)
            
            if not pu.check_files_exist(out_file1,out_file2):
                print("Trimgalore failed")
                return ("",)
            return out_file1,out_file2
            
        else:
            fq=sra_object.localfastqPath
            out_file=os.path.join(out_dir, pu.get_file_basename(fq)+out_suffix+".fastq")
            #giving input arguments as a tuple "--":(fq,)
            newOpts={"--":(fq,),"-o":out_dir}
            #run trimgalore
            mergedOpts={**kwargs,**newOpts}
            
            self.run_trimgalore(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
            """
            running trim galore will create one file named <input>_trimmed.fq
            move these files to the specified out files
            """
            oldFile=os.path.join(out_dir,pu.get_file_basename(fq)+"_trimmed.fq")
            
            pe.move_file(oldFile,out_file)
            
            if not pu.check_files_exist(out_file):
                print("Trimgalore failed")
                return ("",)
            return (out_file,)
        
        
            
    def run_trimgalore(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running trimgalore
        
        Parameters
        ----------
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        
        kwargs: dict
            Options to pass to trimgalore (will override existing parameters)
            
        :return: Status of trimgalore command
        :rtype: bool
        """
        
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
        
        #create command to run
        trimgalore_cmd=['trim_galore']
        trimgalore_cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))
        
        
        #start ececution
        status=pe.execute_command(trimgalore_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("trimgalore failed")
        
        #return status
        return status
        

            
            

class BBmap(RNASeqQC):
    """This class represents bbmap programs
    """
    def __init__(self,**kwargs):
        """
        Parameters
        ----------
        
        kwargs:
            bbduk.sh arguments.
        """
        #run super to inherit parent class properties
        super().__init__() 
        self.programName="bbduk.sh"
        self.dep_list=[self.programName]
        #check if program exists
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
            
        
        self.valid_args=['in','in2','ref','literal','touppercase','interleaved','qin','reads','copyundefined',
                            'samplerate','samref','out','out2','outm','outm2','outs','stats','refstats','rpkm',
                            'dump','duk','nzo','overwrite','showspeed','ziplevel','fastawrap','qout','statscolumns',
                            'rename','refnames','trd','ordered','maxbasesout','maxbasesoutm','','json','bhist','qhist',
                            'qchist','aqhist','bqhist','lhist','phist','gchist','ihist','gcbins','maxhistlen','histbefore',
                            'ehist','qahist','indelhist','mhist','idhist','idbins','varfile','vcf','ignorevcfindels',
                            'k','rcomp','maskmiddle','minkmerhits','minkmerfraction','mincovfraction','hammingdistance',
                            'qhdist','editdistance','hammingdistance2','qhdist2','editdistance2','forbidn','removeifeitherbad',
                            'trimfailures','findbestmatch','skipr1','skipr2','ecco','recalibrate','sam','le.','amino',
                            'threads','prealloc','monitor','minrskip','maxrskip','rskip','qskip','speed','ktrim','kmask',
                            'maskfullycovered','ksplit','mink','qtrim','trimq','trimclip','minlength','mlf','maxlength',
                            'minavgquality','maqb','minbasequality','maxns','mcb','ottm','tp','tbo','strictoverlap',
                            'minoverlap','mininsert','tpe','forcetrimleft','forcetrimright','forcetrimright2',
                            'forcetrimmod','restrictleft','restrictright','mingc','maxgc','gcpairs','tossjunk',
                            'swift','chastityfilter','barcodefilter','barcodes','xmin','ymin','xmax','ymax','trimpolya',
                            'trimpolygleft','trimpolygright','trimpolyg','filterpolyg','pratio','plen','entropy','entropywindow',
                            'entropyk','minbasefrequency','entropytrim','entropymask','entropymark','cardinality',
                            'cardinalityout','loglogk','loglogbuckets','-Xmx','-eoom','-da']
        
        self.passedArgumentDict=kwargs
            
            
            
    def perform_qc(self,sra_object,out_dir="",out_suffix="_bbduk",overwrite=True,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Run bbduk on fastq files specified by the sra_object
        
        Parameters
        ----------
        
        sra_object: SRA
            an SRA object
        out_dir: string
            Path to out dir. Default: sra_object.location
        out_suffix: string
            Suffix for output file name
        overwrite: bool
            overwrite existing files
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        
        kwargs: dict
            options passed to bbduk
            
        :return: Returns the path of fastq files after QC. tuple has one item for single end files and 2 for paired.
        :rtype: tuple
            
        """
        
        #make out_dir
        if not out_dir:
                out_dir=sra_object.location
        else:
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
                    
        if sra_object.layout=='PAIRED':
            fq1=sra_object.localfastq1Path
            fq2=sra_object.localfastq2Path
            
            out_fileName1=pu.get_file_basename(fq1)+out_suffix+".fastq"
            out_fileName2=pu.get_file_basename(fq2)+out_suffix+".fastq"
            out_file1Path=os.path.join(out_dir,out_fileName1)
            out_file2Path=os.path.join(out_dir,out_fileName2)
            
            newOpts={"in":fq1,"in2":fq2,"out":out_file1Path,"out2":out_file2Path}
            mergedOpts={**kwargs,**newOpts}
            
            #run bbduk
            if self.run_bbduk(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts):
                if pu.check_files_exist(out_file1Path,out_file2Path):
                    return(out_file1Path,out_file2Path)
            return("",)
            
            
        else:
            fq=sra_object.localfastqPath
            out_fileName=pu.get_file_basename(fq)+out_suffix+".fastq"
            out_filePath=os.path.join(out_dir,out_fileName)
            newOpts={"in":fq,"out":out_filePath}
            mergedOpts={**kwargs,**newOpts}
            
            #run bbduk
            if self.run_bbduk(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts):
                if pu.check_files_exist(out_filePath):
                    return(out_filePath,)
            return("",)
                    
    
    
    
    
    def run_bbduk(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper to run bbduk.sh
        """
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
        
        #create command to run
        bbduk_cmd=["bbduk.sh"]
        
        #bbduk.sh follows java style arguments
        bbduk_cmd.extend(pu.parse_java_args(self.valid_args,mergedArgsDict))
        
        
        #start ececution
        status=pe.execute_command(bbduk_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("bbduk failed")
        #return status
        return status
    
 
    
    def perform_cleaning(self,sra_object,bbsplit_index,out_dir="",out_suffix="_bbsplit",overwrite=True,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """
        Remove contaminated reads mapping to given reference using bbsplit
        
        Parameters
        ----------
        
        sra_object: SRA
            an SRA object
        bbsplit_index: string
            Path to bbsplit index or fasta file which will generate index
        out_dir: string
            Path to output dir. Default: sra_object.location
        out_suffix: string
            Suffix for output file name
        overwrite: bool
            overwrite existing files
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        
        kwargs: dict
            options passed to bbsplit

            :return: Returns the path of fastq files after QC. tuple has one item for single end files and 2 for paired.
            :rtype: tuple
        """
        
        #check index
        indexPath=""
        if not pu.check_paths_exist(bbsplit_index):
            #index folder doesn't exist
            #check if input is path to fasta
            if not pu.check_files_exist(bbsplit_index):
                print("Error: Please check bbsplit index")
                return ("",)
            #check if index folder "ref" exists in this directory
            indexPath=os.path.join(pu.get_file_directory(bbsplit_index),"ref")
            if pu.check_paths_exist(indexPath):
                print("Using bbsplit index: "+indexPath)
            else:
                #create new index
                print("Creating new index"+indexPath)
                newOpts={"ref_x":bbsplit_index,"path": pu.get_file_directory(bbsplit_index)}
                mergedOpts={**kwargs,**newOpts}
                #run bbduk
                if not self.run_bbsplit(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts):
                    print("Error creating bbsplit index.")
                    return ("",)
                if not pu.check_paths_exist(indexPath):
                    print("Error creating bbsplit index.")
                    return ("",)
        else:
            indexPath=bbsplit_index
                
        
        #indexPath point to the ref directory, go one directory higher
        indexPath=os.path.dirname(indexPath)
        
        
        #make out_dir
        if not out_dir:
                out_dir=sra_object.location
        else:
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
        
        if sra_object.layout=='PAIRED':
            fq1=sra_object.localfastq1Path
            fq2=sra_object.localfastq2Path
            #append input and output options
            
            out_fileName1=pu.get_file_basename(fq1)+out_suffix+".fastq"
            out_fileName2=pu.get_file_basename(fq2)+out_suffix+".fastq"
            out_file1Path=os.path.join(out_dir,out_fileName1)
            out_file2Path=os.path.join(out_dir,out_fileName2)
            
            newOpts={"in1":fq1,"in2":fq2,"outu1":out_file1Path,"outu2":out_file2Path,"path":indexPath}
            mergedOpts={**kwargs,**newOpts}
            
            #run bbsplit
            if self.run_bbsplit(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts):
                if pu.check_files_exist(out_file1Path,out_file2Path):
                    return(out_file1Path,out_file2Path)
            return("",)
            
            
        else:
            fq=sra_object.localfastqPath
            #append input and output options
           
            out_fileName=pu.get_file_basename(fq)+out_suffix+".fastq"
            out_filePath=os.path.join(out_dir,out_fileName)
            newOpts={"in":fq,"outu":out_filePath,"path":indexPath}
            mergedOpts={**kwargs,**newOpts}
            
            #run bbsplit
            if self.run_bbsplit(verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts):
                if pu.check_files_exist(out_filePath):
                    return(out_filePath,)
            
            return("",)
    
    
    
    def run_bbsplit(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """wrapper to run bbsplit
        
        :return: Status of bbsplit command
        :rtype: bool
        """
        
        bbsplit_args=['ref','ref_x','build','path','in','in1','in2','outu','outu2','outu1','qin','interleaved',
                          'maxindel','minratio','minhits','ambiguous','ambiguous2',
                          'qtrim','untrim','out_','basename','bs','scafstats',
                          'refstats','nzo','-Xmx','-eoom','-da']
        
        #override existing arguments
        #don't use class arguments
        mergedArgsDict={**kwargs}
        
        #create command to run
        bbsp_cmd=["bbsplit.sh"]
        
        #bbduk.sh follows java style arguments
        bbsp_cmd.extend(pu.parse_java_args(bbsplit_args,mergedArgsDict))
        
        
        #start ececution
        status=pe.execute_command(bbsp_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("bbsplit failed")
        #return status
        return status
    
    
    
    
    
    
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov 24 19:53:42 2019

@author: usingh
contains classes of RNA-Seq alignment programs
"""

from pyrpipe import pyrpipe_utils as pu
from pyrpipe import pyrpipe_engine as pe
import os

class Aligner:
    """This is an abstract class for alignment programs.
    """
    def __init__(self,index=""):
        self.category="Aligner"
        self.passedArgumentDict={}
        self.index=index
        
    def build_index(self):
        """function to create an index used by the aligner
        """
        pass
    
    def check_index(self):
        """Function to check if index of this object is valid and exists
        """
    
    def perform_alignment(self,sra_object):
        """Function to perform alignment taking and sraobject as input
        
        """
        pass

class Hisat2(Aligner):
    """This class represents hisat2 program.
    
       Parameters
       ----------       
       
       hisat2_index: string
            path to the histat2 index. This index is stored with the object and will be used when hisat is invoked using this object.
            
       **kwargs: dict
            parameters passed to the hisat2 program. These parameters could be overridden later when running hisat.
            
    Attributes
    ----------
    
    """ 
    def __init__(self,hisat2_index="",**kwargs):
        
        super().__init__() 
        self.programName="hisat2"
        #check if hisat2 exists
        if not pe.check_dependencies([self.programName]):
            raise Exception("ERROR: "+ self.programName+" not found.")
        
        self.valid_args=['-x','-1','-2','-U','--sra-acc','-S','-q','--qseq','-f','-r','-c','-s',
                            '-u','-5','-3','--phred33','--phred64','--int-quals',
                            '--sra-acc','--n-ceil','--ignore-quals','--nofw','--norc','--pen-cansplice',
                            '--pen-noncansplice','--pen-canintronlen','--pen-noncanintronlen','--min-intronlen'
                            ,'--max-intronlen','--known-splicesite-infile','--novel-splicesite-outfile',
                            '--novel-splicesite-infile','--no-temp-splicesite','--no-spliced-alignment',
                            '--rna-strandness','--tmo','--dta','--dta-cufflinks','--avoid-pseudogene',
                            '--no-templatelen-adjustment','--mp','--sp','--no-softclip','--np','--rdg',
                            '--rfg','--score-min','-k','-I','-X','--fr','--rf','--ff','--no-mixed',
                            '--no-discordant','-t','--un','--al','--un-conc','--al-conc','--un-gz',
                            '--summary-file','--new-summary','--quiet','--met-file','--met-stderr',
                            '--met','--no-head','--no-sq','--rg-id','--rgit-sec-seq','-o','-p',
                            '--reorder','--mm','--qc-filter','--seed','--non-deterministic',
                            '--remove-chrname','--add-chrname','--version']
        
        
        #initialize the passed arguments
        self.passedArgumentDict=kwargs
        
        #if index is passed, update the passed arguments
        if len(hisat2_index)>0 and pu.check_hisatindex(hisat2_index):
            print("HISAT2 index is: "+hisat2_index)
            self.hisat2_index=hisat2_index
            self.passedArgumentDict['-x']=self.hisat2_index
            self.index=self.hisat2_index
        else:
            print("No Hisat2 index provided. Please build index now to generate an index using build_Index()....")
            
        
        
            
    def build_index(self,index_path,index_name,*args,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Build a hisat index with given parameters and saves the new index to self.hisat2_index.
        
        Parameters
        ----------
        
        index_path: string
            Path where the index will be created
            
        index_name: string
            A name for the index
            
        args: tuple
            Path to reference input files
            
        verbose : bool
            Print stdout and std error
            
        quiet : bool 
            Print nothing
            
        logs : bool 
            Log this command to pyrpipe logs
            
        objectid : string 
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        
        kwargs: dict
            Parameters for the hisat2-build command
        
        
            
        :return: Returns the status of hisat2-build
        :rtype: bool
        """
        
        #check input references
        if len(args)<1:
            pu.print_boldred("No reference sequence provided to hisat2-build. Exiting")
            return False
        
        if not pu.check_files_exist(*args):
            pu.print_boldred("Please check input reference sequences provided to hisat2-build. Exiting")
            return False
            
        overwrite=True
        print("Building hisat index...")
        
        hisat2Buildvalid_args=['-c','--large-index','-a','-p','--bmax','--bmaxdivn','--dcv','--nodc','-r','-3','-o',
                                  '-t','--localoffrate','--localftabchars','--snp','--haplotype','--ss','--exon',
                                  '--seed','-q','-h','--usage','--version']
        #create the out dir
        if not pu.check_paths_exist(index_path):
            if not pu.mkdir(index_path):
                print("ERROR in building hisat2 index. Failed to create index directory.")
                return False
        
        if not overwrite:
            #check if files exists
            if pu.check_hisatindex(os.path.join(index_path,index_name)):
                print("Hisat2 index with same name already exists. Exiting...")
                return True
        
        hisat2Build_Cmd=['hisat2-build']
        #add options
        hisat2Build_Cmd.extend(pu.parse_unix_args(hisat2Buildvalid_args,kwargs))
        #add input files
        hisat2Build_Cmd.append(str(",".join(args)))
        #add dir/basenae
        hisat2Build_Cmd.append(os.path.join(index_path,index_name))
        #print("Executing:"+str(" ".join(hisat2Build_Cmd)))
        
        #start ececution
        status=pe.execute_command(hisat2Build_Cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("hisatBuild failed")
            return False
        
        #check index files
        if not pu.check_hisatindex(os.path.join(index_path,index_name)):
            pu.print_boldred("hisatBuild failed")
            return False
        
        #set the index path
        self.hisat2_index=os.path.join(index_path,index_name)
        self.passedArgumentDict['-x']=self.hisat2_index
        
        #return status
        return True
        
        
    def perform_alignment(self,sra_object,out_suffix="_hisat2",verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to perform alignment using sra_object.
        
        Parameters
        ----------
        
        sra_object SRA object
            An object of type SRA. The path to fastq files will be obtained from this object.
        out_suffix: string
            Suffix for the output sam file
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to hisat2. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).
        """
        
        
        #create path to output sam file
        outSamFile=os.path.join(sra_object.location,sra_object.srr_accession+out_suffix+".sam")
        
        """
        Handle overwrite
        """
        overwrite=True
        if not overwrite:
            #check if file exists. return if yes
            if os.path.isfile(outSamFile):
                print("The file "+outSamFile+" already exists. Exiting..")
                return outSamFile
        
        #find layout and fq file paths
        if sra_object.layout == 'PAIRED':
            newOpts={"-1":sra_object.localfastq1Path,"-2":sra_object.localfastq2Path,"-S":outSamFile}
        else:
            newOpts={"-U":sra_object.localfastqPath,"-S":outSamFile}
        
        #add input files to kwargs, overwrite kwargs with newOpts
        mergedOpts={**kwargs,**newOpts}
        
        #call run_hisat2
        status=self.run_hisat2(verbose=verbose,quiet=quiet,logs=logs,objectid=sra_object.srr_accession,**mergedOpts)
        
        if status:
            #check if sam file is present in the location directory of sra_object
            if pu.check_files_exist(outSamFile):
                return outSamFile
        else:
            return ""
            
        
    def run_hisat2(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running hisat2.
        
        Parameters
        ----------
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        
        kwargs: dict
            arguments to pass to hisat2. This will override parametrs already existing in the self.passedArgumentList list but NOT replace them.

        :return: Returns the status of hisat2. True is passed, False if failed.
        :rtype: bool
        """
        
        #check for a valid index
        if not self.check_index():
            raise Exception("ERROR: Invalid HISAT2 index. Please run build index to generate an index.")
            
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
       
        hisat2_Cmd=['hisat2']
        #add options
        hisat2_Cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))        
        
        #execute command
        cmd_status=pe.execute_command(hisat2_Cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not cmd_status:
            print("hisat2 failed:"+" ".join(hisat2_Cmd))
     
        #return status
        return cmd_status
        
        
    
    def check_index(self):
        if hasattr(self,'hisat2_index'):
            return(pu.check_hisatindex(self.hisat2_index))
        else:
            return False

    



class Star(Aligner):
    """This class represents STAR program.
    
       Parameters
       ----------
       
       star_index: string
            path to a star index. This index will be used when star is invoked using this object.
       kwargs: dict
            parameters passed to the star program. These parameters could be overridden later when running star.
            
    Attributes
    ----------
    """ 
    def __init__(self,star_index="",**kwargs):
        
        super().__init__() 
        self.programName="STAR"
        
        self.dep_list=[self.programName]        
        #check if star exists
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
            
        self.valid_args=['--help','--parametersFiles','--sysShell','--runMode','--runThreadN','--runDirPerm','--runRNGseed','--quantMode','--quantTranscriptomeBAMcompression','--quantTranscriptomeBan','--twopassMode','--twopass1readsN',
                            '--genomeDir','--genomeLoad','--genomeFastaFiles','--genomeChrBinNbits','--genomeSAindexNbases','--genomeSAsparseD','--genomeSuffixLengthMax','--genomeChainFiles','--genomeFileSizes',
                            '--sjdbFileChrStartEnd','--sjdbGTFfile','--sjdbGTFchrPrefix','--sjdbGTFfeatureExon','--sjdbGTFtagExonParentTranscript','--sjdbGTFtagExonParentGene','--sjdbOverhang','--sjdbScore','--sjdbInsertSave',
                            '--inputBAMfile','--readFilesIn','--readFilesCommand','--readMapNumber','--readMatesLengthsIn','--readNameSeparator','--clip3pNbases','--clip5pNbases','--clip3pAdapterSeq','--clip3pAdapterMMp','--clip3pAfterAdapterNbases',
                            '--limitGenomeGenerateRAM','--limitIObufferSize','--limitOutSAMoneReadBytes','--limitOutSJoneRead','--limitOutSJcollapsed','--limitBAMsortRAM ','--limitSjdbInsertNsj','--outFileNamePrefix','--outTmpDir','--outTmpKeep',
                            '--outStd','--outReadsUnmapped','--outQSconversionAdd','--outMultimapperOrder','--outSAMtype','--outSAMmode','--outSAMstrandField','--outSAMattributes','--outSAMattrIHstart','--outSAMunmapped','--outSAMorder',
                            '--outSAMprimaryFlag','--outSAMreadID','--outSAMmapqUnique','--outSAMflagOR','--outSAMflagAND','--outSAMattrRGline','--outSAMheaderHD','--outSAMheaderPG','--outSAMheaderCommentFile','--outSAMfilter','--outSAMmultNmax',
                            '--outBAMcompression','--outBAMsortingThreadN','--bamRemoveDuplicatesType','--bamRemoveDuplicatesMate2basesN','--outWigType','--outWigStrand','--outWigReferencesPrefix','--outWigNorm','--outFilterType',
                            '--outFilterMultimapScoreRange','--outFilterMultimapNmax','--outFilterMismatchNmax','--outFilterMismatchNoverLmax','--outFilterMismatchNoverReadLmax','--outFilterScoreMin','--outFilterScoreMinOverLread',
                            '--outFilterMatchNmin','--outFilterMatchNminOverLread','--outFilterIntronMotifs','--outSJfilterReads','--outSJfilterOverhangMin','--outSJfilterCountUniqueMin','--outSJfilterCountTotalMin','--outSJfilterDistToOtherSJmin',
                            '--outSJfilterIntronMaxVsReadN','--scoreGap','--scoreGapNoncan','--scoreGapGCAG ','--scoreGapATAC','--scoreGenomicLengthLog2scale','--scoreDelOpen','--scoreDelBase','--scoreInsOpen','--scoreInsBase','--scoreStitchSJshift',
                            '--seedSearchStartLmax','--seedSearchStartLmaxOverLread','--seedSearchLmax','--seedMultimapNmax','--seedPerReadNmax','--seedPerWindowNmax','--seedNoneLociPerWindow','--alignIntronMin','--alignIntronMax','--alignMatesGapMax',
                            '--alignSJoverhangMin','--alignSJstitchMismatchNmax','--alignSJDBoverhangMin','--alignSplicedMateMapLmin','--alignSplicedMateMapLminOverLmate','--alignWindowsPerReadNmax','--alignTranscriptsPerWindowNmax','--alignTranscriptsPerReadNmax',
                            '--alignEndsType','--alignEndsProtrude','--alignSoftClipAtReferenceEnds','--winAnchorMultimapNmax','--winBinNbits','--winAnchorDistNbins','--winFlankNbins','--winReadCoverageRelativeMin','--winReadCoverageBasesMin',
                            '--chimOutType','--chimSegmentMin','--chimScoreMin','--chimScoreDropMax','--chimScoreSeparation','--chimScoreJunctionNonGTAG','--chimJunctionOverhangMin','--chimSegmentReadGapMax','--chimFilter','--chimMainSegmentMultNmax']
                
       
        #initialize the passed arguments
        self.passedArgumentDict=kwargs
        
        #if index is passed, update the passed arguments
        if len(star_index)>0 and pu.check_starindex(star_index):
            print("STAR index is: "+star_index)
            self.star_index=star_index
            self.passedArgumentDict['--genomeDir']=self.star_index
        else:
            print("No STAR index provided. Please build index now to generate an index using build_index()....")
            
    
    
    def build_index(self,index_path,*args,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Build a star index with given parameters and saves the new index to self.star_index.
        
        Parameters
        ----------
        
        index_path: string
            Path where the index will be created
        
        args: tuple
            Path to reference input files
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        
        kwargs: dict
            Parameters for the star command

        :return: Returns status of star command
        :rtype: bool
        """
        
        #if index already exists then exit
        if pu.check_starindex(index_path):
            pu.print_green("STAR index already exists. Using it...")
            self.star_index=index_path
            self.passedArgumentDict['--genomeDir']=self.star_index
            return True
            
        
        #check input files
        if len(args)<1:
            pu.print_boldred("Please provide input fasta file to build STAR index")
            return False
        
        if not pu.check_files_exist(*args):
            raise Exception("Please check input to star index")
            return False
        
        #create path if doesnt exist
        if not pu.check_paths_exist(index_path):
            if not pu.mkdir(index_path):
                raise Exception("Error creating STAR index. Exiting.")
                return False
        
        #add runMode
        newOpts={"--runMode":"genomeGenerate","--genomeDir":index_path,"--genomeFastaFiles":" ".join(args)}
        
        mergedOpts={**kwargs,**newOpts}
        
        starbuild_Cmd=['STAR']
        starbuild_Cmd.extend(pu.parse_unix_args(self.valid_args,mergedOpts))
        
        #execute command
        status=pe.execute_command(starbuild_Cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        
        
        if status:
            if pu.check_paths_exist(index_path):
                #update object's index
                self.star_index=index_path
                self.passedArgumentDict['--genomeDir']=self.star_index
                if self.check_index():
                    return True
        else:
            return False
        
 
            
    def perform_alignment(self,sra_object,out_suffix="_star",out_dir="",verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to perform alignment using star and the provided SRA object.
        All star output will be written to the sra_object directory by default.
        
        Parameters
        ----------
        
        sra_object: SRA object
            An object of type SRA. The path to fastq files will be obtained from this object.
        out_suffix: string
            Suffix for the output file
        out_dir: str
            outout directory default: sra_object.location
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to STAR. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).
        

        :return: Return the path to the output dir
        :rtype: string
        """
        
        if not out_dir:
            out_dir=sra_object.location
        else:
            #create out_dir if not exists
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
        
        #find layout and fq file paths
        if sra_object.layout == 'PAIRED':
            newOpts={"--readFilesIn":sra_object.localfastq1Path+" "+sra_object.localfastq2Path}
        else:
            newOpts={"--readFilesIn":sra_object.localfastqPath}
        
        #add out dir
        newOpts["--outFileNamePrefix"]=out_dir+"/"
        
               
        #add input files to kwargs, overwrite kwargs with newOpts
        mergedOpts={**kwargs,**newOpts}
        
        #call star
        status=self.run_star(verbose=verbose,quiet=quiet,logs=logs,objectid=sra_object.srr_accession,**mergedOpts)
                
        
        if status:
            print("Star finished")
            if pu.check_paths_exist(out_dir):
                return out_dir
        else:
            return ""
        
    
    def run_star(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running star.
        The self.star_index index used.
        
        Parameters
        ----------
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to stringtie. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).
        kwargs: dict
            arguments to pass to star. This will override parametrs already existing in the self.passedArgumentList list but NOT replace all of them.
            
        :return: Returns the status of star. True is passed, False if failed.
        :rtype: bool
        """
        
        #check for a valid index
        if not self.check_index():
            raise Exception("ERROR: Invalid star index. Please run build index to generate an index.")
            
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
       
        star_cmd=['STAR']
        #add options
        star_cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))        
        
        #execute command
        cmd_status=pe.execute_command(star_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        
        if not cmd_status:
            print("STAR failed:"+" ".join(star_cmd))
     
        #return status
        return cmd_status
    
    
    def check_index(self):
        if hasattr(self,'star_index'):
            return(pu.check_starindex(self.star_index))
        else:
            return False
            



class Bowtie2(Aligner):
    """This class represents bowtie2 program.
    
       Parameters
       ----------
       
       bowtie2_index: string
            path to a bowtie2 index. This index will be used when bowtie2 is invoked using this object.
       **kwargs: dict
            parameters passed to the bowtie2 program. These parameters could be overridden later when running bowtie2.
            
       Attributes
       ----------
    """ 
    def __init__(self,bowtie2_index,**kwargs):
        """Bowtie2 constructor. Initialize bowtie2 index and other parameters.
        """       
        
        super().__init__() 
        self.programName="bowtie2"
        self.dep_list=[self.programName]        
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
        
        self.valid_args=['-x','-1','-2','-U','--interleaved','-S','-b','-q','--tab5','--tab6','--qseq','-f','-r','-F','-c','-s','-u','-5','-3',
                            '--trim-to','--phred33','--phred64','--int-quals','--very-fast','--fast',
                            '--sensitive','--very-sensitive','--very-fast-local','--fast-local',
                            '--sensitive-local','--very-sensitive-local','-N','-L','-i','--n-ceil',
                            '--dpad','--gbar','--ignore-quals','--nofw','--norc','--no-1mm-upfront',
                            '--end-to-end','--local','--ma','--mp','--np','--rdg','--rfg','--score-min',
                            '-k','-a','-D','-R','-I','-X','--fr','--rf','--ff','--no-mixed','--no-discordant',
                            '--dovetail','--no-contain','--no-overlap','--align-paired-reads','--preserve-tags',
                            '-t','--un','--al','--un-conc','--al-conc','--un-gz','--quiet','--met-file',
                            '--met-stderr','--met','--no-unal','--no-head','--no-sq','--rg-id','--rg',
                            '--omit-sec-seq','--sam-no-qname-trunc','--xeq','--soft-clipped-unmapped-tlen',
                            '-p','--threads','--reorder','--mm','--qc-filter','--seed','--non-deterministic',
                            '--version','-h','--help']
        
        #initialize the passed arguments
        self.passedArgumentDict=kwargs
        
        #if index is passed, update the passed arguments
        if len(bowtie2_index)>0 and pu.check_bowtie2index(bowtie2_index):
            print("Bowtie2 index is: "+bowtie2_index)
            self.bowtie2_index=bowtie2_index
            self.passedArgumentDict['-x']=self.bowtie2_index
        else:
            print("No Bowtie2 index provided. Please build index now to generate an index...")
        
        
    def build_index(self,index_path,index_name,*args,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Build a bowtie2 index with given parameters and saves the new index to self.bowtie2_index.
        
        Parameters
        ----------
        
        index_path: string
            Path where the index will be created
        index_name: string
            A name for the index
        arg3: tuple
            Path to reference input files
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to bowtie2. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).
            
        
 
        :return: Returns the status of bowtie-build
        :rtype: bool
        """
        
        #check input references
        if len(args)<1:
            pu.print_boldred("No reference sequence provided to bowtie2-build. Exiting")
            return False
        
        if not pu.check_files_exist(*args):
            pu.print_boldred("Please check input reference sequences provided to bowtie2-build. Exiting")
            return False
            
        overwrite=True
        
        
        bowtie2_build_args=['-f','-c','--large-index','--debug','--sanitized','--verbose','-a',
                            '--noauto','-p','--packed','--bmax','--bmaxdivn','--dcv','--nodc',
                            '-r','--noref','-3','--justref','-o','--offrate','-t','--ftabchars',
                            '--threads','--seed','-q','--quiet','-h','--help','--usage','--version']
        #create the out dir
        if not pu.check_paths_exist(index_path):
            if not pu.mkdir(index_path):
                print("ERROR in building bowtie2 index. Failed to create index directory.")
                return False
        
        if not overwrite:
            #check if files exists
            if pu.check_bowtie2index(os.path.join(index_path,index_name)):
                print("bowtie2 index with same name already exists. Exiting...")
                return True
            
        bowtie2Build_Cmd=['bowtie2-build']
        #add options
        bowtie2Build_Cmd.extend(pu.parse_unix_args(bowtie2_build_args,kwargs))
        #add input files
        bowtie2Build_Cmd.append(str(",".join(args)))
        #add dir/basenae
        bowtie2Build_Cmd.append(os.path.join(index_path,index_name))
        #print("Executing:"+str(" ".join(hisat2Build_Cmd)))
        
        #start ececution
        status=pe.execute_command(bowtie2Build_Cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("bowtie2-build failed")
            return False
        
        #check index files
        if not pu.check_bowtie2index(os.path.join(index_path,index_name)):
            pu.print_boldred("bowtie2-build failed")
            return False
        
        #set the index path
        self.bowtie2_index=os.path.join(index_path,index_name)
        self.passedArgumentDict['-x']=self.bowtie2_index
        
        #return status
        return True
        
    
    def perform_alignment(self,sra_object,out_suffix="_bt2",out_dir="",overwrite=True,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to perform alignment using self object and the provided sra_object.
        
        Parameters
        ----------
        
        sra_object: SRA object
            An object of type SRA. The path to fastq files will be obtained from this object.
        out_suffix: string
            Suffix for the output sam file
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to bowtie2. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).
        
        :return: Returns the output sam file path
        :rtype: string
        """
        if not out_dir:
            out_dir=sra_object.location
        else:
            if not pu.check_paths_exist(out_dir):
                pu.mkdir(out_dir)
                
        #create path to output sam file
        outFile=os.path.join(out_dir,sra_object.srr_accession+out_suffix+".sam")
                    
        """
        Handle overwrite
        """
        overwrite=True
        if not overwrite:
            #check if file exists. return if yes
            if os.path.isfile(outFile):
                print("The file "+outFile+" already exists. Exiting..")
                return outFile
        
        #find layout and fq file paths
        if sra_object.layout == 'PAIRED':
            newOpts={"-1":sra_object.localfastq1Path,"-2":sra_object.localfastq2Path,"-S":outFile}
        else:
            newOpts={"-U":sra_object.localfastqPath,"-S":outFile}
        
        #add input files to kwargs, overwrite kwargs with newOpts
        mergedOpts={**kwargs,**newOpts}
        
        status=self.run_bowtie2(verbose=verbose,quiet=quiet,logs=logs,objectid=sra_object.srr_accession,**mergedOpts)
        
        if status:
            #check if sam file is present in the location directory of sra_object
            if pu.check_files_exist(outFile):
                return outFile
        else:
            return ""
        
        
        
    
    def run_bowtie2(self,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running bowtie2.
        
        
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        
        kwargs: dict
            arguments to pass to bowtie2. This will override parametrs already existing in the self.passedArgumentList list but NOT replace them.
            
        :return: Returns the status of bowtie2. True is passed, False if failed.
        :rtype: bool
        """
        
        #check for a valid index
        if not self.check_index():
            raise Exception("ERROR: Invalid Bowtie2 index. Please run build index to generate an index.")
        
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
            
        bowtie2_cmd=['bowtie2']
        bowtie2_cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))
        
        #print("Executing:"+" ".join(bowtie2_cmd))
        
        #start ececution
        status=pe.execute_command(bowtie2_cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid)
        if not status:
            pu.print_boldred("bowtie2 failed")
        return status
    
    
    def check_index(self):
        """Function to check bowtie index.
        Returns True is index exist on disk.
        """
        if hasattr(self,'bowtie2_index'):
            return(pu.check_bowtie2index(self.bowtie2_index))
        return False



         



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 10 13:40:01 2019

@author: usingh
"""
import dill
import datetime as dt
import os
#importing pyrpipe_engine here causes issues and stalling of log
#from pyrpipe import pyrpipe_engine as pre
#import pyrpipe

def getTimestamp(shorten=False):
    """Return timestamp YYYYMMDDHHMISE
    shorten: return shorter version without spaces.
    """
    timestamp=str(dt.datetime.now()).split(".")[0].replace(" ","-")
    if shorten:
        timestamp=timestamp.replace("-","").replace(" ","").replace(":","")
    return timestamp

def save_session(filename,add_timestamp=True,out_dir=""):
    """Save current workspace using dill.
    Returns True is save is successful
    """
    #timestamp format YYYYMMDDHHMISE
    timestamp=getTimestamp(True)
    
    
    if not out_dir:        
        out_dir=os.getcwd()
    else:
        if not os.path.isdir(out_dir):
            os.makedirs(out_dir)
    
    outFile=os.path.join(out_dir,filename)
    if add_timestamp:
        outFile=outFile+"_"+timestamp+".pyrpipe"
    else:
        outFile=outFile+".pyrpipe"
    
    """
    Do not pickle logger. This causes problems when restoring session with python < 3.7
    Delete all logger instances. 
    del pre.pyrpipeLoggerObject
    del pyrpipe.pyrpipe_engine.pyrpipeLoggerObject
    """    
    """
    creating a logger class fixed this issue
    """ 
    
    
    #save workspace
    try:
        dill.dump_session(outFile)
        print("Session saved to: "+outFile)
        return True
    except Exception:
        raise Exception("Failed to save session")
        return False


def restore_session(file):
    """Resore a session from file.
    
    :return: Returns True if session is restored
    :rtype: bool
    """
    if not os.path.isfile(file):
        print(file+" doesn't exist")
        return False
    #load the session
    dill.load_session(file)
    print("Session restored.")
    return True


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 21 16:58:13 2019

@author: usingh
"""

from pyrpipe import pyrpipe_utils as pu
import datetime as dt
import json
import seaborn as sns
import pandas as pd
import copy 
import matplotlib.pyplot as plt

import os

class Benchmark:
    """Class to generate benchmark reports from pyrpipe logs.
    
    Parameters
    ----------
    
    log_file: string
        path to the log file
    env_log: string
        path to the ENV log file
    out_dir: string
        path to the output directory
        
        
    """
    def __init__(self,log_file,env_log,out_dir=""):
        
        
        if not pu.check_files_exist(log_file,env_log):
            raise Exception("Please check input for benchmark report. {} {}".format(log_file,env_log))
        if not out_dir:
            out_dir=os.getcwd()
        self.log_file=log_file
        self.env_log=env_log
        self.runtimes_by_prog={}
        self.runtimes_by_object={}
        #init
        pu.print_blue("parsing log...")
        self.parse_logs()
        pu.print_blue("done.")
        #out_dir
        self.benchmark_dir=os.path.join(out_dir,'benchmark_reports')
        if not pu.check_paths_exist(self.benchmark_dir):
            if not pu.mkdir(self.benchmark_dir):
                raise Exception("Error running benchmarks. Can not create output directory {}".format(self.benchmark_dir))
            
        
        
    def parse_runtime(self,timestring):
        """
        Parse runtime as string and return seconds.
        
        Returns: float
            runtime in sec
        """
        try:
            runtime= dt.datetime.strptime(timestring,"%H:%M:%S")
            delta_time = dt.timedelta(days=0,hours=runtime.hour, minutes=runtime.minute, seconds=runtime.second)
            return delta_time.seconds
        except ValueError:
            #try days format
            temp=timestring.split(",")
            days=int(temp[0].split(" ")[0].strip())
            rest=temp[1].strip()
            #hours=int(days)*24
            runtime= dt.datetime.strptime(rest,"%H:%M:%S")
            #one day less
            #lastruntime=lastruntime+dt.timedelta(days=days-1)
            delta_time = dt.timedelta(days=days,hours=runtime.hour, minutes=runtime.minute, seconds=runtime.second)
            return delta_time.seconds+(86400*days)
        
    def parse_logs(self):
        """Parse the input logs.
        For each command create a dict with runtimes as list and program name as key
        For each object id in input log create a dict.
        The followind dicts are created:
        
        runtimes_by_prog contains runtimes for each program. program is the key and the runtimes are in a list in order as they apprear in the log file.
        
        runtimes_by_object is nested a dict containing runtimes for each object by each program. e.g. {'ob1':{'prog1':[1,2,3],'prog2':[1,2,3]}, 'ob2':{'prog1':[12,22,13],'prog2':[1,2,3]} }
        """
        
        
        
        with open(self.log_file) as f:
            data=f.read().splitlines()
        num_commands=0

    
        for l in data:
            if not l.startswith("#"):
                thisDict=json.loads(l)
                num_commands+=1
                #ignore failed commands
                if int(thisDict['exitcode'])!=0:
                    continue
                             
                #store runtimes by programname
                try:
                    programname=thisDict['commandname']
                except KeyError:
                    #for older logs
                    programname=thisDict['cmd'].split(" ")[0]
                runtime=self.parse_runtime(thisDict['runtime'])
                #add to dict
                if programname in self.runtimes_by_prog:
                    self.runtimes_by_prog[programname].append(runtime)
                else:
                    self.runtimes_by_prog[programname]=[runtime]
                    
                #store runtimes by object id
                try:
                    objectid=thisDict['objectid']
                except KeyError:
                    objectid='SRR'+str(num_commands%50)
                
                if objectid in self.runtimes_by_object:
                    #if the object is used with same program extend the list
                    if programname in self.runtimes_by_object[objectid]:
                        self.runtimes_by_object[objectid][programname].append(runtime)
                    else:
                        self.runtimes_by_object[objectid][programname]=[runtime]
                else:
                    self.runtimes_by_object[objectid]={programname:[runtime]}
                
        #print(self.runtimes_by_prog)
        #print(self.runtimes_by_object)
                                   
    def get_time_perobject(self,func="sum"):
        """Returns a dataframe containing total execution time for each object in a pyrpipe log.
        An object is identified by the objectid e.g. SRR accession.
        """
        result=pd.DataFrame()
        for k in self.runtimes_by_object:
            
            v=copy.deepcopy(self.runtimes_by_object[k]) #don't change the self.runtimes_by_object
            
            totaltime=0
            for prog in v:
                runtimes=v[prog]                
                if func == "mean":
                    valtoput=sum(runtimes)/len(runtimes)
                    totaltime+=valtoput
                    v[prog]=[valtoput]
                else:                    
                    valtoput=sum(runtimes)
                    totaltime+=valtoput
                    v[prog]=[valtoput]
                    
            #add current id
            v['id']=k
            v['total']=totaltime
            
            result=result.append(pd.DataFrame.from_dict(v),sort=False)
            
        return result
        
    def plot_time_perobject(self):
        """Function to plot charts summarizing runtimes for each object in the pipeline.
        The charts are save to the out_dir path.
        """
        data=self.get_time_perobject()
        #remove rows with no object id
        data=data[data['id']!='NA']
        cols=list(data.columns)
        id_col=cols.index('id')
        #swap id with 0
        temp=cols[0]
        cols[0]='id'
        cols[id_col]=temp
        total_col=cols.index('total')
        #swap id with last
        temp=cols[-1]
        cols[-1]='total'
        cols[total_col]=temp
        
        data=data[cols]
        
        total_rows=data.shape[0]
        sns.set_context('poster')
        f, ax = plt.subplots(figsize = (total_rows/4,total_rows/2))
        
        sns.set_color_codes('bright')
        sns.barplot(x = 'total', y = 'id', data = data, label = 'Total', color = 'black', edgecolor = 'w')
        
        
        
        
        #for each col
        current_palette = sns.color_palette("colorblind")
        i=0
        for col in data.columns:
            if col in ['id','total']:
                continue
            sns.barplot(x = col, y = 'id', data = data,label = col, color = current_palette[i], edgecolor = 'w',)
            i+=1

        #add legend
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        #ax.legend(ncol = 2, loc = 'upper right')
        sns.despine(left = True, bottom = True)
        ax.set(xlabel='runtime (sec.)')
        #plt.show()
        
        #save plot
        plotfile=os.path.join(self.benchmark_dir,'time_per_object.png')
        plt.savefig(plotfile,bbox_inches='tight')
        
        #write data to out_dir
        outfile=os.path.join(self.benchmark_dir,'time_per_object.csv')
        data.to_csv(outfile, index=False)
    
    
    def get_time_perprogram(self):
        """Returns a dataframe with program execution times.
        """
        result=pd.DataFrame()
        for k in self.runtimes_by_prog:
            v=self.runtimes_by_prog[k]
            total=sum(v)
            mean=sum(v)/len(v)
            row={'program':k,'total':[total],'average':[mean]}
            #add row to dataframe
            result=result.append(pd.DataFrame.from_dict(row),sort=False)
        return result
    
    def get_programtime_boxdata(self):
        """Return dataframe to make box plot of program times
        """
        box_data=pd.DataFrame({ key:pd.Series(value) for key, value in self.runtimes_by_prog.items() })
        #return box_data
        
        
        ndf=pd.DataFrame(columns=['data','name'])
        for i in range(box_data.shape[1]):
            to_append=list(box_data.iloc[:,i])
            col_name=box_data.columns[i]
            col_list=[col_name for x in to_append]
            for j in range(len(to_append)):
                ndf=ndf.append({'data':to_append[j],'name':col_list[j]},ignore_index=True )
        
        return ndf
        
        
    def plot_time_perprogram(self):
        """Function to plot charts to summarize runtimes of each program
        """
        data=self.get_time_perprogram()
        sns.set_context('poster')
        f, ax = plt.subplots()
        
        current_palette = sns.color_palette("colorblind")
        
        #plot total time
        sns.barplot(x = 'total', y = 'program', data = data, label = 'Total', color = current_palette[4])
        #plot mean time
        sns.barplot(x = 'average', y = 'program', data = data, label = 'Avg.', color = current_palette[2])
        
        #save the barplot
        #add legend
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        #ax.legend(ncol = 2, loc = 'upper left')
        sns.despine(left = True, bottom = True)
        ax.set(xlabel='runtime (sec.)')
        #save plot
        plotfile=os.path.join(self.benchmark_dir,'time_per_program.png')
        plt.savefig(plotfile,bbox_inches='tight')
        
        
        #clear plot
        plt.clf()
        #make boxplots for runtime distribution
        sns.set_color_codes('bright')
        sns.set_context('poster')
        box_data=self.get_programtime_boxdata()
        #convert data to floats for boxplot
        box_data['data']=box_data['data'].astype(float)
        #print(box_data)
        numprog=len(self.runtimes_by_prog.keys())
        #sns.set(style="ticks")
        # Initialize the figure with a logarithmic x axis
        f, ax = plt.subplots(figsize=(20, numprog*2))
        ax.set_xscale("log")
        
        #sns.boxplot( data=box_data,orient="h")
        sns.boxplot(x="data", y="name", data=box_data)
        # Add in points to show each observation
        sns.swarmplot(x="data", y="name", data=box_data,size=5, color=".3", linewidth=0)
        
        ax.xaxis.grid(True)
        ax.set(ylabel="")
        ax.set(xlabel="runtime (sec.)")
        sns.despine(left=True)
        f.suptitle('Boxplot showing runtimes of each program', fontsize=26)
        
        
        #save to file
        plotfile=os.path.join(self.benchmark_dir,'program_boxplots.png')
        plt.savefig(plotfile,bbox_inches='tight')
        #plt.show()
        #clear plot
        
        plt.clf()
        
        
        #make pie charts
        sns.set_color_codes('bright')
        sns.set_context('poster')
        f, ax = plt.subplots(figsize=(10, numprog*2))
        current_palette = sns.color_palette("colorblind")
        plt.pie(data['total'], colors=current_palette, labels= data['program'],counterclock=False, shadow=False,autopct='%1.1f%%')
        f.suptitle('Summary of total runtimes', fontsize=26)
        
        #save plot
        plotfile=os.path.join(self.benchmark_dir,'program_summary.png')
        plt.savefig(plotfile,bbox_inches='tight')
        
        #write data to out_dir
        outfile=os.path.join(self.benchmark_dir,'time_per_program.csv')
        data.to_csv(outfile, index=False)
        #save boxplot data
        box_data=pd.DataFrame({ key:pd.Series(value) for key, value in self.runtimes_by_prog.items() })
        outfile=os.path.join(self.benchmark_dir,'program_box_data.csv')
        box_data.to_csv(outfile, index=False)
        
        
            
            





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan  2 18:20:41 2020

@author: usingh
"""

from pyrpipe import pyrpipe_utils as pu
from pyrpipe import pyrpipe_engine as pe
import os

class Quant:
    """This is an abstract class for quantification programs.
    """
    def __init__(self,index=""):
        self.category="Quantification"
        self.passedArgumentDict={}
        self.index=index
        
    def build_index(self):
        """function to create an index used by the quantification program
        """
        pass
    
    def check_index(self):
        """Function to check if index of this object is valid and exists
        """
    
    def perform_quant(self,sra_object):
        """Function to perform quant taking and sraobject as input
        
        """
        pass
    
class Kallisto(Quant):
    """This class represents kallisto
    
    kallisto_index: string
        path to kallisto index
    kwargs: dict
        parametrs passed to kallisto
    """
    
    def __init__(self,kallisto_index,**kwargs):
        super().__init__() 
        self.programName="kallisto"
        self.dep_list=[self.programName]        
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
        
        
        ##kallisto index
        self.validArgsIndex=['-i','--index','-k','--kmer-size','--make-unique']
        ##kallisto quant
        self.validArgsQuant=['-i','--index','-o','--output-dir','--bias','-b','--bootstrap-samples',
                             '--seed','--plaintext','--fusion','--single','--fr-stranded','--rf-stranded',
                             '-l','--fragment-length','-s','--sd','-t','--threads','--pseudobam']
        ##kallisto pseudo
        self.validArgsPseudo=['-i','--index','-o','--output-dir','-u','--umi','-b','--batch',
                              '--single','-l','--fragment-length','-s','--sd','-t','--threads','--pseudobam']
            ##kallisto h5dump
        self.validArgsh5dump=['-o','--output-dir']
        
        self.valid_args=pu.get_union(self.validArgsIndex,self.validArgsQuant,self.validArgsPseudo,self.validArgsh5dump)
        
        #initialize the passed arguments
        self.passedArgumentDict=kwargs
        
        #if index is passed, update the passed arguments
        if len(kallisto_index)>0 and pu.check_files_exist(kallisto_index):
            print("kallisto index is: "+kallisto_index)
            self.kallisto_index=kallisto_index
            self.passedArgumentDict['-i']=self.kallisto_index
        else:
            print("No kallisto index provided. Please use build_index() now to generate an index...")
            
    def build_index(self,index_path,index_name,fasta,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Function to  build kallisto index
        
        index_path: str
            path to the output directory
        index_name: str
            index name
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to kallisto. This will override the existing options in self.passed_args_dict (only replace existing arguments and not replace all the arguments).
            
        :return: Status of kallisto index
        :rtype: bool
        """
        
        #check input
        if not pu.check_files_exist(fasta):
            pu.print_boldred("{} does not exist. Exiting".format(fasta))
            return False
        
        #create out dir
        if not pu.check_paths_exist(index_path):
            if not pu.mkdir(index_path):
                print("ERROR in building kallisto index. Failed to create index directory.")
                return False
            
        indexOut=os.path.join(index_path,index_name)
        newOpts={"--":(fasta,),"-i":indexOut}
        mergedOpts={**kwargs,**newOpts}
        
        #call salmon
        status=self.run_kallisto("index",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if status:
            #check if sam file is present in the location directory of sra_object
            if pu.check_files_exist(indexOut):
                self.kallisto_index=indexOut
                self.passedArgumentDict['-i']=self.kallisto_index
                pu.print_green("kallisto_index is:"+self.kallisto_index)
                return True
        else:
            pu.print_boldred("Failed to create kallisto index")
            return False
    
    def perform_quant(self,sra_object,out_dir="",verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Run kallisto quant
        
        sra_object: SRA
            SRA object contatining paths to fastq files
        index_path: str
            path to the output directory
        index_name: str
            index name
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to kallisto. This will override the existing options

        :return: Path to kallisto out directory
        :rtype: string
        """
        
        if not out_dir:
            out_dir=os.path.join(sra_object.location,"kallisto_out")
        
        
        
        if sra_object.layout == 'PAIRED':
            newOpts={"-o":out_dir,"--":(sra_object.localfastq1Path,sra_object.localfastq2Path)}
        else:
            newOpts={"-o":out_dir,"--single":"", "--":(sra_object.localfastqPath,)}
        
        
        #add input files to kwargs, overwrite kwargs with newOpts
        mergedOpts={**kwargs,**newOpts}
        
        #call salmon
        status=self.run_kallisto("quant",verbose=verbose,quiet=quiet,logs=logs,objectid=sra_object.srr_accession,**mergedOpts)
        
        if status:
            #check if sam file is present in the location directory of sra_object
            if pu.check_files_exist(os.path.join(out_dir,"abundance.tsv")):
                return out_dir
        
        pu.print_boldred("kallisto quant failed")
        return ""
        
    
    
    def run_kallisto(self,subcommand,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running kallisto.
        
        Parameters
        ----------
        
        subcommand: str
            subcommand for kallisto
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to kallisto. This will override the existing options

        :return: Returns the status of kallisto. True is passed, False if failed.
        :rtype: bool
        """
        
        #check for a valid index
        if subcommand!="index":
            if not self.check_index():
                raise Exception("ERROR: Invalid kallisto index. Please run build index to generate an index.")
        
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
            
        kallisto_Cmd=['kallisto',subcommand]
        kallisto_Cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))
        
        #start ececution
        status=pe.execute_command(kallisto_Cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,command_name=" ".join(kallisto_Cmd[0:2]))
        if not status:
            pu.print_boldred("kallisto failed")
        return status       
    
    def check_index(self):
        """Check valid kallisto index
        """
        if hasattr(self,'kallisto_index'):
            return(pu.check_files_exist(self.kallisto_index))
        return False
            


class Salmon(Quant):
    """This class represents salmon
    
    salmon_index: string
        Path to salmon index
    kwargs: dict
        Options passed to salmon
    """      
    def __init__(self,salmon_index,**kwargs):    
        super().__init__() 
        self.programName="salmon"
        self.dep_list=[self.programName]        
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: "+ self.programName+" not found.")
        
        
        ##salmon index
        self.validArgsIndex=['-v','--version','-h','--help','-t','--transcripts','-k','--kmerLen','-i',
                             '--index','--gencode','--keepDuplicates','-p','--threads','--perfectHash',
                             '--type','-s','--sasamp']
        ##salmon quant read
        self.validArgsQuantReads=['--help-reads','-i','--index','-l','--libType','-r','--unmatedReads',
                                  '-1','--mates1','-2','--mates2','-o','--output','--discardOrphansQuasi',
                                  '--allowOrphansFMD','--seqBias','--gcBias','-p','--threads','--incompatPrior',
                                  '-g','--geneMap','-z','--writeMappings','--meta','--alternativeInitMode',
                                  '--auxDir','-c','--consistentHits','--dumpEq','-d','--dumpEqWeights',
                                  '--fasterMapping','--minAssignedFrags','--reduceGCMemory','--biasSpeedSamp',
                                  '--strictIntersect','--fldMax','--fldMean','--fldSD','-f','--forgettingFactor',
                                  '-m','--maxOcc','--initUniform','-w','--maxReadOcc','--noLengthCorrection',
                                  '--noEffectiveLengthCorrection','--noFragLengthDist','--noBiasLengthThreshold',
                                  '--numBiasSamples','--numAuxModelSamples','--numPreAuxModelSamples','--useVBOpt',
                                  '--rangeFactorizationBins','--numGibbsSamples','--numBootstraps','--thinningFactor',
                                  '-q','--perTranscriptPrior','--vbPrior','--writeOrphanLinks','--writeUnmappedNames',
                                  '-x','--quasiCoverage']
        ##salmon quant alignment
        self.validArgsQuantAlign=['--help-alignment','-l','--libType','-a','--alignments','-t','--targets','-p',
                                  '--threads','--seqBias','--gcBias','--incompatPrior','--useErrorModel',
                                  '-o','--output','--meta','-g','--geneMap','--alternativeInitMode','--auxDir'
                                  ,'--noBiasLengthThreshold','--dumpEq','-d','--dumpEqWeights','--fldMax',
                                  '--fldMean','--fldSD','-f','--forgettingFactor','--minAssignedFrags',
                                  '--gencode','--reduceGCMemory','--biasSpeedSamp','--mappingCacheMemoryLimit',
                                  '-w','--maxReadOcc','--noEffectiveLengthCorrection','--noFragLengthDist',
                                  '-v','--useVBOpt','--rangeFactorizationBins','--perTranscriptPrior','--vbPrior',
                                  '--numErrorBins','--numBiasSamples','--numPreAuxModelSamples','--numAuxModelSamples',
                                  '-s','--sampleOut','-u','--sampleUnaligned','--numGibbsSamples','--numBootstraps',
                                  '--thinningFactor']
        ##salmon quantmerge
        self.validArgsQuantMerge=['--quants','--names','-c','--column','-o','--output']

        self.valid_args=pu.get_union(self.validArgsIndex,self.validArgsQuantReads,self.validArgsQuantAlign,self.validArgsQuantMerge)
        
        #initialize the passed arguments
        self.passedArgumentDict=kwargs
        
        #if index is passed, update the passed arguments
        if len(salmon_index)>0 and pu.check_salmon_index(salmon_index):
            print("salmon index is: "+salmon_index)
            self.salmon_index=salmon_index
            self.passedArgumentDict['-i']=self.salmon_index
        else:
            print("No salmon index provided. Please build index now to generate an index...")
            
            
            
    def build_index(self,index_path,index_name,fasta,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """
        build salmon index and store the path to index in self
        
        index_path: str
            path to the output directory
        index_name: str
            index name
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to salmon. This will override the existing options
            
        :return: status of salmon index
        :rtype: bool
        """
        
        #check input
        if not pu.check_files_exist(fasta):
            pu.print_boldred("{} does not exist. Exiting".format(fasta))
            return False
        #create out dir
        if not pu.check_paths_exist(index_path):
            if not pu.mkdir(index_path):
                print("ERROR in building hisat2 index. Failed to create index directory.")
                return False
        indexOut=os.path.join(index_path,index_name)
        newOpts={"-t":fasta,"-i":indexOut}
        mergedOpts={**kwargs,**newOpts}
        
        #call salmon
        status=self.run_salmon("index",verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,**mergedOpts)
        
        if status:
            #check if sam file is present in the location directory of sra_object
            #if check_files_exist(os.path.join(indexOut,"versionInfo.json")): #not sure if this is reliable
            if pu.check_paths_exist(indexOut):
                self.salmon_index=indexOut
                self.passedArgumentDict['-i']=self.salmon_index
                pu.print_green("salmon index is:"+self.salmon_index)
                return True
        
        pu.print_boldred("Failed to create salmon index")
        return False
        
        
    
    def perform_quant(self,sra_object,out_dir="",libType="A",verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """run salmon quant
        
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to salmon. This will override the existing options

        :return: Path to salmon out directory
        :rtype: string
        """
        
        if not out_dir:
            out_dir=os.path.join(sra_object.location,"salmon_out")
        
        
        
        if sra_object.layout == 'PAIRED':
            newOpts={"-o":out_dir,"-l":libType,"-1":sra_object.localfastq1Path,"-2":sra_object.localfastq2Path}
        else:
            newOpts={"-o":out_dir,"-l":libType,"-r":sra_object.localfastqPath}
        
        
        #add input files to kwargs, overwrite kwargs with newOpts
        mergedOpts={**kwargs,**newOpts}
        
        #call salmon
        status=self.run_salmon("quant",verbose=verbose,quiet=quiet,logs=logs,objectid=sra_object.srr_accession,**mergedOpts)
        
        if status:
            #check if sam file is present in the location directory of sra_object
            if pu.check_files_exist(os.path.join(out_dir,"quant.sf")):
                return out_dir
        
        pu.print_boldred("salmon quant failed")
        return ""
        
        
        
    def run_salmon(self,subcommand,verbose=False,quiet=False,logs=True,objectid="NA",**kwargs):
        """Wrapper for running salmon.
        
        Parameters
        ----------
        
        subcommand: str
            subcommand for salmon
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        objectid: str
            Provide an id to attach with this command e.g. the SRR accession. This is useful for debugging, benchmarking and reports.
        kwargs: dict
            Options to pass to salmon. This will override the existing options

        :return: Returns the status of salmon. True is passed, False if failed.
        :rtype: bool
        """
        
        #check for a valid index
        if subcommand!="index":
            if not self.check_index():
                raise Exception("ERROR: Invalid salmon index. Please run build index to generate an index.")
        
        #override existing arguments
        mergedArgsDict={**self.passedArgumentDict,**kwargs}
            
        salmon_Cmd=['salmon',subcommand]
        salmon_Cmd.extend(pu.parse_unix_args(self.valid_args,mergedArgsDict))
        
        #start ececution
        status=pe.execute_command(salmon_Cmd,verbose=verbose,quiet=quiet,logs=logs,objectid=objectid,command_name=" ".join(salmon_Cmd[0:2]))
        if not status:
            pu.print_boldred("salmon failed")
        return status 

    def check_index(self):
        if hasattr(self,'salmon_index'):
            return pu.check_salmonindex(self.salmon_index)
        return False
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 23 15:45:26 2019

@author: usingh


"""

from pyrpipe import pyrpipe_utils as pu
from pyrpipe import pyrpipe_engine as pe
import os


class SRA:
    """This class represents an SRA object
    
        Parameters
        ----------
        
        srr_accession: string
            A valid SRR accession
        location: string
            Path where all data related to this object (e.g. .sra files, metadata, fastq files) will be stored.
            Default value of the path will be "./<SRR_accession>". <SRR_accession> is added at the end of the path
            so that final location is location/<SRR_accession>.
            For consistency, location and SRR Accession id are not allowed to be modified.
            
        scan_path: string
            If RNA-Seq data already exists locally, provide the scan path to scan a directory and create an SRA object.
        
        Attributes
        -----------
        
        """
    def __init__(self,srr_accession=None, location=None):
        
        if location is None and srr_accession is None:
            #pu.print_boldred("Please provide a valid srr accession or location, or both")
            raise Exception("Please provide a valid srr accession or location, or both")
            #use the scan_path to create sra object
            #self.init_from_path(scan_path)
        #else:
            #use the provided srr_accession
        self.init_from_accession(srr_accession,location)
        
        ##check if sra, fastq files already exists
    
    def init_from_path(self,path):
        if not pu.check_paths_exist(path):
            raise Exception("Please provide a valid path to scan for RNA-Seq data")
        
        #scan path for fastq
        self.search_fastq(path)
        #scan path for sra
        self.search_sra(path)
        if not (self.fastqFilesExistsLocally() or self.sraFileExistsLocally()):
                raise Exception("No files found at:"+ path+ "Please provide a valid path to scan for RNA-Seq data")
    
    def search_sra(self,path):
        """Search .sra file under a dir
        Return True if found otherwise False
        """
        #search files under the path
        
        sra_files=pe.find_files(path,"*.sra")
        
        if len(sra_files)<1:
            return False
        
        if len(sra_files)>1:
            pu.print_boldred("Found multiple .sra files. Using the first entry...")
        sra_path=sra_files[0]
        #self.location=path
        self.srr_accession=pu.get_file_basename(sra_path)
        self.localSRAFilePath=sra_path
        self.sraFileSize=pu.get_file_size(self.localSRAFilePath)
        #test if file is paired or single end
        if pe.is_paired(self.localSRAFilePath):
            self.layout="PAIRED"
        else:
            self.layout="SINGLE"
        
        pu.print_green("Found .sra "+self.localSRAFilePath)
        return True
        
    def search_fastq(self,path):
        """Search .fastq file under a dir and create SRA object
        Return True if found otherwise False
        """
        #search files under the path
        fq_files=pe.find_files(path,"*.fastq")
        
        if len(fq_files)<1:
            return False
        
        if len(fq_files)>2:
            pu.print_boldred("Can not determine .fastq. Exiting...")
            return False
        
        fq_files.sort()
        #case with single fastq
        if len(fq_files)==1:
            self.localfastqPath=fq_files[0]
            pu.print_green("Found .fastq "+self.localfastqPath)
            self.layout="SINGLE"
        
        #case with paired fastq
        if len(fq_files)==2:
            self.localfastq1Path=fq_files[0]
            self.localfastq2Path=fq_files[1]
            pu.print_green("Found .fastq "+self.localfastq1Path+" "+self.localfastq2Path)
            self.layout="PAIRED"
        
        #self.location=path
        #self.srr_accession=pu.get_file_basename(fq_files[0])
        return True
        
    
    def init_from_accession(self,srr_accession,location):
        """Create SRA object using provided srr accession and location to save the data
        """
        self.dep_list=['prefetch',"fasterq-dump"]
        if not pe.check_dependencies(self.dep_list):
            raise Exception("ERROR: Please install missing programs.")
        
        if srr_accession is None:
            raise Exception("Please provide a valid accession")
            
        if location is None:
            location=os.getcwd()
        #pu.print_info("Creating SRA: "+srr_accession)
        self.srr_accession=srr_accession
        #create a dir named <srr_accession> and use as location
        self.location=os.path.join(location,self.srr_accession)
    
        #search for existing files in location
        #self.search_fastq(self.location)
        #scan path for sra
        #self.search_sra(self.location)
        
        
        #check SRA file
        if pu.check_files_exist(os.path.join(self.location,self.srr_accession+".sra")):
            pu.print_green(self.srr_accession+".sra exists.")
            self.localSRAFilePath=os.path.join(self.location,self.srr_accession+".sra")
            self.sraFileSize=pu.get_file_size(self.localSRAFilePath)
            #test if file is paired or single end
            if pe.is_paired(self.localSRAFilePath):
                self.layout="PAIRED"
            else:
                self.layout="SINGLE"
        
        #check fastq file
        self.search_fastq(self.location)
        
        
        
        
    def __setattr__(self, name, value):
        """Make srr accession immutable
        """
        immutableFields=['srr_accession','location']
        
        if (name in immutableFields and name in self.__dict__):
            raise Exception("Can not modify "+name)
        else:
            self.__dict__[name] = value
                
    
    def download_fastq(self,verbose=False,quiet=False,logs=True,procs=2,**kwargs):
        """Function to download fastq files
        """
        
        #check if fastq files exists already
        if self.fastqFilesExistsLocally():
            pu.print_green("Fastq files exist already")
            return True
        
        
        fasterqdumpArgsList=['-f','-t','-s','-N','-X','-a','-p','-c','-o','-O','-h','-V',
                             '-L','-v','-q','-b','-m','-x','-S','-3','-P','-M',
                             '-B','--option-file','--strict','--table','--include-technical',
                             '--skip-technical','--concatenate-reads']
        fstrqd_Cmd=['fasterq-dump']
        fstrqd_Cmd.extend(pu.parse_unix_args(fasterqdumpArgsList,kwargs))
        #add location
        fstrqd_Cmd.extend(['-O',self.location])
        #add output filename. output will be <srr_accession>.fastq or <srr_accession>_1.fastq and <srr_accession>_2.fastq
        fstrqd_Cmd.extend(['-o',self.srr_accession+".fastq"])
        fstrqd_Cmd.extend(['-e',str(procs)])
        if self.sraFileExistsLocally():
            fstrqd_Cmd.append(self.localSRAFilePath)
        else:
            fstrqd_Cmd.append(self.srr_accession)
        
        #execute command
        cmdStatus=pe.execute_command(fstrqd_Cmd,objectid=self.srr_accession)
        if not cmdStatus:
            print("fasterqdump failed for:"+self.srr_accession)
            return False        
        
        
        if not hasattr(self,'layout'):
            fq_files=pe.find_files(self.location,self.srr_accession+"*.fastq")
            if len(fq_files)==1:
                self.layout='SINGLE'
            else:
                self.layout='PAIRED'
        
        #check if fastq files are downloaded        
        if(self.layout=="SINGLE"):
            self.localfastqPath=os.path.join(self.location,self.srr_accession+".fastq")
            if not pu.check_files_exist(self.localfastqPath):
                pu.print_boldred("Error running fasterq-dump file. File "+self.localfastqPath+" does not exist!!!")
                return False
        else:
            self.localfastq1Path=os.path.join(self.location,self.srr_accession+"_1.fastq")
            self.localfastq2Path=os.path.join(self.location,self.srr_accession+"_2.fastq")
            if not pu.check_files_exist(self.localfastq1Path,self.localfastq2Path):
                pu.print_boldred("Error running fasterq-dump file. File "+self.localfastq1Path+" does not exist!!!")
                return False
        
            
            
            
        return True
        
        
  
    def download_sra(self,verbose=False,quiet=False,logs=True,**kwargs):
        """This function downloads .sra file from NCBI SRA servers using the prefetch command.

        NCBI sra-toolkit 2.9 or higher must be installed on the system in order to use prefetch. 
        prefetch will create a folder with name same as <srr_accession> under the location (path) specified.
        The path of downloaded file is saved in the object as localSRAPath. This localSRAPath is then used
        by other functions to access the downloaded data. 
        The **kwargs is for passing arguments to the prefetch command.
        
        Parameters
        ----------
        
        kwargs: dict
            dict containing additional prefetch arguments

        :return: Return status of the prefetch command. True if successful download and False if failed.
        :rtype: bool

        Examples
        --------
        >>> object.download_sra()
        True
        """
        
        
        #store path to the downloaded sra file
        self.localSRAFilePath=os.path.join(self.location,self.srr_accession+".sra")
        #check if already exists
        if pu.check_files_exist(self.localSRAFilePath):
            pu.print_green("File already exists:"+self.localSRAFilePath)
            #save file .sra file size
            self.sraFileSize=pu.get_file_size(self.localSRAFilePath)
            #test if file is paired or single end
            if pe.is_paired(self.localSRAFilePath):
                self.layout="PAIRED"
            else:
                self.layout="SINGLE"
            return True
            
        
        pu.print_info("Downloading "+self.srr_accession+" ...")
        
        #scan for prefetch arguments
        prefetchArgsList=['-f','-t','-l','-n','-s','-R','-N','-X','-o','-a','--ascp-options','-p','--eliminate-quals','-c','-o','-O','-h','-V','-L','-v','-q']
        
        
        #ignore location and file name arguments if given
        if '-O' in kwargs:
            print("Ignoring -O flag."+" location is: "+self.location)
            #delete -O parameter
            del kwargs['-O']
        if '-o' in kwargs:
            print("Ignoring -o flag."+" File name is: "+self.srr_accession)
            #delete -o parameter
            del kwargs['-o']
            

        prefetch_Cmd=['prefetch']
        prefetch_Cmd.extend(pu.parse_unix_args(prefetchArgsList,kwargs))
        prefetch_Cmd.extend(['-O',self.location])
        prefetch_Cmd.append(self.srr_accession)
                
        cmdStatus=pe.execute_command(prefetch_Cmd,objectid=self.srr_accession)
        if not cmdStatus:
            pu.print_boldred("prefetch failed for:"+self.srr_accession)
            return False
        
        
        #validate path exists
        if not pu.check_files_exist(self.localSRAFilePath):
            pu.print_boldred("Error downloading file. File "+self.localSRAFilePath+" does not exist!!!")
            return False
        
        print ("Downloaded file: "+self.localSRAFilePath+" {0} ".format(pu.get_file_size(self.localSRAFilePath)))
        #save file .sra file size
        self.sraFileSize=pu.get_file_size(self.localSRAFilePath)
        #test if file is paired or single end
        if pe.is_paired(self.localSRAFilePath):
            self.layout="PAIRED"
        else:
            self.layout="SINGLE"
            
        
        return True
    
    
           
    
    def sraFileExistsLocally(self):
        """Function to check if sra file is present on disk
        """
        if hasattr(self,'localSRAFilePath'):
            return(os.path.isfile(self.localSRAFilePath))
        else:
            return False
        
    def fastqFilesExistsLocally(self):
        """Function to check if fastq file is present on disk
        """
        
        if not hasattr(self,'layout'):
            return False
        
        if self.layout=='PAIRED':
            if hasattr(self,'localfastq1Path'):
                if not pu.check_files_exist(self.localfastq1Path):
                    return False
            else:
                return False
            
            if hasattr(self,'localfastq2Path'):
                if not pu.check_files_exist(self.localfastq2Path):
                    return False
            else:
                return False
            
            return True
            
        else:
            if hasattr(self,'localfastqPath'):
                return pu.check_files_exist(self.localfastqPath)
            else:            
                return False
    
    def run_fasterqdump(self,delete_sra=False,verbose=False,quiet=False,logs=True,**kwargs):
        """Execute fasterq-dump to convert .sra file to fastq files.
        The fastq files will be stored in the same directory as the sra file. All fastq files should be consistently named
        using the extension .fastq
        
        Parameters
        ----------
        
        delete_sra: bool
            delete sra file after completion
        verbose: bool
            Print stdout and std error
        quiet: bool
            Print nothing
        logs: bool
            Log this command to pyrpipe logs
        kwargs: dict
            A dict containing fasterq-dump arguments
        
        :return: Return status of the fasterq-dump command. True if successful download and False if failed.
        :rtype: bool

        Examples
        --------
        >>> object.run_fasterqdump()
        True
        """
        #check if fastq files exists already
        if self.fastqFilesExistsLocally():
            pu.print_green("Fastq files exist already")
            return True
        
        #first check is sra exists
        if not self.sraFileExistsLocally():
            pu.print_boldred("Error executing fasterq-dump: .sra file not found. Please run download_sra().")
            return False
        #else directly run fasterq-dump on accession ?
        
        fasterqdumpArgsList=['-f','-t','-s','-N','-X','-a','-p','-c','-o','-O','-h','-V',
                             '-L','-v','-q','-b','-m','-e','-x','-S','-3','-P','-M',
                             '-B','--option-file','--strict','--table','--include-technical',
                             '--skip-technical','--concatenate-reads']
        
        
        
        #ignore location and file name arguments if given
        if '-O' in kwargs:
            print("Ignoring -O flag."+" location is: "+self.location)
            #delete -O parameter
            del kwargs['-O']
        if '-o' in kwargs:
            print("Ignoring -o flag."+" File name is: "+self.srr_accession)
            #delete -o parameter
            del kwargs['-o']
        
        
        #execute command
        
        fstrqd_Cmd=['fasterq-dump']
        fstrqd_Cmd.extend(pu.parse_unix_args(fasterqdumpArgsList,kwargs))
        #add location
        fstrqd_Cmd.extend(['-O',self.location])
        #add output filename. output will be <srr_accession>.fastq or <srr_accession>_1.fastq and <srr_accession>_2.fastq
        fstrqd_Cmd.extend(['-o',self.srr_accession+".fastq"])
        fstrqd_Cmd.append(self.localSRAFilePath)
        
        #execute command
        cmdStatus=pe.execute_command(fstrqd_Cmd,objectid=self.srr_accession)
        if not cmdStatus:
            print("fasterqdump failed for:"+self.srr_accession)
            return False
        
        
        #check if fastq files are downloaded 
        if(self.layout=="SINGLE"):
            self.localfastqPath=os.path.join(self.location,self.srr_accession+".fastq")
            
            if not pu.check_files_exist(self.localfastqPath):
                pu.print_boldred("Error running fasterq-dump file. File "+self.localfastqPath+" does not exist!!!")
                return False
            
        else:
            self.localfastq1Path=os.path.join(self.location,self.srr_accession+"_1.fastq")
            self.localfastq2Path=os.path.join(self.location,self.srr_accession+"_2.fastq")
            
            if not pu.check_files_exist(self.localfastq1Path,self.localfastq2Path):
                pu.print_boldred("Error running fasterq-dump file. File "+self.localfastq1Path+" does not exist!!!")
                return False
            
        #delete sra file if specified
        if delete_sra:
            self.delete_sra()
            
        return True
    
    def perform_qc(self,qcObject,deleteRawFastq=False):
        """Function to perform quality control with specified qc object.
        A qc object refers to one of the RNA-Seq qc program like trim_galore oe bbduk.
        The qcObject should be initialized with all the parameters.
        By default the trimmed/qc fastq files will be generated in the same directory as the original fastq files.
        After QC, this SRA object will update the localfastqPath or localfastq1Path and localfastq2Path variables to store the new fastq files.
        New variables localRawfastqPath or localRawfastq1Path and localRawfastq2Path will be created to store the paths of original fastq files.
      
        Parameters
        ----------
        
        qcObject: RNASeqQC object
            qcObject specifying the program to be used. The object contains the necessary parametrs to execute the parameters
            
        deleteRawFastq: bool
            Delete the raw fastq files after QC

        :return: Return status of the QC. True if successful download and False if failed.
        :rtype: bool

        Examples
        --------
        >>> object.perform_qc(qc.BBmap())
        True
        """
        #check a valid qcObject
        if not (hasattr(qcObject,'category') and qcObject.category=='RNASeqQC'):
            print ("Error: No valid QC object provided. Skipping QC for "+self.srr_accession)
            return False
        
        #save thq qc object for later references
        self.QCObject=qcObject
        print("Performing QC using "+qcObject.programName)
        #each qcObject has a function run() to execute their method
        qcStatus=qcObject.perform_qc(self,objectid=self.srr_accession)
        
        #if job failed
        if not qcStatus[0]:
            print ("Error performing QC for "+self.srr_accession)
            return False
        
        if self.layout=='PAIRED':
            
            #delete old fastq files if specified
            if deleteRawFastq:
                self.delete_fastq()
            
            else:
                #create new fields to refer to older fastq files
                self.localRawfastq1Path=self.localfastq1Path
                self.localRawfastq2Path=self.localfastq2Path
            
            ##update local fastq path
            self.localfastq1Path=qcStatus[0]    
            self.localfastq2Path=qcStatus[1]
        else:
            if deleteRawFastq:
                self.delete_fastq()
            else:
                self.localRawfastqPath=self.localfastqPath
                
            self.localfastqPath=qcStatus[0]
        
        
        
        return True
    
    def delete_fastq(self):
        """Delte the fastq files from the disk.
        The files are referenced by self.localfastqPath or self.localfastq1Path and self.localfastq2Path
        """
        if self.layout=='PAIRED':
            if pe.deleteMultipleFilesFromDisk(self.localfastq1Path,self.localfastq2Path):
                del self.localfastq1Path
                del self.localfastq2Path
                return True
        else:
            if pe.deleteFileFromDisk(self.localfastqPath):
                del self.localfastqPath
                return True
        return False
        
        
    def delete_sra(self):
        """Delete the downloaded SRA files.
        """
        if(pe.deleteFileFromDisk(self.localSRAFilePath)):
            del self.localSRAFilePath
            return True
        return False
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 10 11:16:40 2019

@author: usingh
"""

import sys
import os
import argparse
import json
from pyrpipe import pyrpipe_utils as pu
from jinja2 import Environment, BaseLoader
from weasyprint import HTML,CSS
from html import escape
import datetime as dt
import multiqc as mc
from pyrpipe import benchmark as bm


try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from pyrpipe import report_templates


def parseEnvLog(envLog):
    #parse the env log
    with open(envLog) as f:
        envdata=f.read().splitlines()
    
    sysInfo={}
    progList={}
    for l in envdata:
        if not l.startswith("#"):
            if not sysInfo:
                sysInfo=json.loads(l)
            else:
                thisProgram=json.loads(l)
                progList[thisProgram['name']]=thisProgram
    return sysInfo,progList

def generateEnvReportTable(sysInfo,progList):
    """create html table to list environment
    Parameters
    ----------
    sysInfo: dict
        system information from env log file
    progList: dict
        programs and their information from env log
    """
    
    
    
    os=sysInfo['os']
    now=sysInfo['now']
    sysmodulesList=sysInfo['sysmodules'].strip('][').split(', ')
    syspathList=sysInfo['syspath'].strip('][').split(', ')
    python=sysInfo['python']
    cpu=sysInfo['cpu']
    
    ##create system info table
    tabStr='\n<h2>Environment Information</h2>'
    tabStr+='\n <div class="envtabs">'
    #add program Info
    tabStr+='\n<table class="programinfo" >'
    tabStr+='<tr><th colspan="3">Programs</th></tr>'
    tabStr+='\n<tr> <th>{}</th> <th>{}</th> <th>{}</th>  </tr>'.format("name","version","path")
    for k, v in progList.items():
        tabStr+='\n<tr> <td>{}</td> <td>{}</td> <td>{}</td>  </tr>'.format(v["name"],v["version"],v["path"])
    tabStr+='\n</table>'    
    tabStr+='\n<br><br>'
    tabStr+='\n<table class="sysInfotable" >'
    tabStr+='<tr><th colspan="2">System Information</th></tr>'
    tabStr+='\n<tr>    <td>Time at collection</td> <td>{}</td>    </tr>'.format(now)
    tabStr+='\n<tr>    <td>Python</td> <td>{}</td>    </tr>'.format(python)
    tabStr+='\n<tr>    <td>Operating system</td> <td>{}</td>    </tr>'.format(os)
    tabStr+='\n<tr>    <td>CPU</td> <td>{}</td>    </tr>'.format(cpu)
    tabStr+='\n</table>'
    tabStr+='\n<br><br>'
    #add sys modules table
    tabStr+='\n<table class="sysmodules" >'
    tabStr+='<tr><th colspan="1">sys.modules</th></tr>'
    for s in sysmodulesList:
        tabStr+='\n<tr><td>{}</td></tr>'.format(s)
    tabStr+='\n</table>'
    tabStr+='\n<br><br>'
    #add sys path table
    tabStr+='\n<table class="syspath" >'
    tabStr+='<tr><th colspan="1">sys.path</th></tr>'
    for s in syspathList:
        tabStr+='\n<tr><td>{}</td></tr>'.format(s)
    tabStr+='\n</table>'
    tabStr+='\n<br><br>'
    tabStr+='\n</div>'
            
    
    
    return tabStr

    
    
def generateHTMLReport(templateFile,cmdLog,envLog,coverage='f'):
    """Generates html report
    Parameters
    ----------
    templatefile: string
        path to a template file
    cmdlog: string
        path to the log file
    envlog: string
        path to the env log file
    coverage: string
        tpye of report: full, summary, fail, pass
    
    """
    #vars for generating summary
    startTime=""
    endTime=""
    numCommands=0
    failedCommands=0
    passedCommands=0
    numPrograms=0
    progNames=[]
    #parse envLog
    sysInfo,progList=parseEnvLog(envLog)
    
    #get starttime #end time is calculated from log below
    startTime=dt.datetime.strptime(sysInfo['now'],"%y-%m-%d %H:%M:%S")
    #total progs used
    progNames=progList.keys()
    numPrograms=len(progNames)
    
    #read the template
    templateHTMLFile = pkg_resources.read_text(report_templates, templateFile)
    #load template file
    template = Environment(loader=BaseLoader()).from_string(templateHTMLFile)
    
    """
    for each command executed get the report parameters.
    Template should specify the parameters:
        cmd: the command executed
        stdout: the stdout
        stderr: the stderr
        exitcode: return code obtained
        starttime: time started
        runtime: execution time
        
        example record:{'cmd':logMessage,
                 'exitcode':str(exitCode),
                 'runtime':str(dt.timedelta(seconds=timeDiff)),
                 'starttime':str(strStartTime),
                 'stdout':stdout,
                 'stderr':stderr                 
                }
    """
    with open(cmdLog) as f:
        data=f.read().splitlines()
    
    #read head.html
    headHTML=pkg_resources.read_text(report_templates, 'head.html')
    #add file name
    headHTML+='\n<h2> <em>pyrpipe</em> report</h2>'
    headHTML+='\n<stron>file name:{}</strong>'.format(cmdLog)
    headHTML+='\n<hr><br><br>\n'
    
    fullHTML="\n<h2> Details </h2>"
    failColor="rgb(208,28,139)"
    passColor="rgb(77,172,38)" 
    for l in data:
        if not l.startswith("#"):
            thisDict=json.loads(l)
            numCommands+=1
            #add color to table
            if int(thisDict['exitcode'])==0:
                thisDict['statuscolor']=passColor
                passedCommands+=1
            else:
                thisDict['statuscolor']=failColor
                failedCommands+=1
            
            #program name
            programname=thisDict['commandname']
            #if programname == "":
            #    programname=thisDict['cmd'].split(" ")[0]
            #add program version info
            newDict={**thisDict,**progList[programname]}
            
            #skip passed
            if coverage=='i' and int(thisDict['exitcode'])==0:
                continue
            #skip failed
            if coverage=='p' and int(thisDict['exitcode'])!=0:
                continue
            
            #escape all special html charecters
            for k, v in newDict.items():
                newDict[k] = escape(str(v))
            fullHTML=fullHTML+"\n"+template.render(newDict)
            
    #get start and runtime of last command
    lastDict=json.loads(data[-1])
    lastST=dt.datetime.strptime(lastDict['starttime'],"%y-%m-%d %H:%M:%S")
    try:
        lastruntime= dt.datetime.strptime(lastDict['runtime'],"%H:%M:%S")
        deltaTime = dt.timedelta(days=0,hours=lastruntime.hour, minutes=lastruntime.minute, seconds=lastruntime.second)
    except ValueError:
        #try days format
        timeString=lastDict['runtime'].split(",")
        days=int(timeString[0].split(" ")[0].strip())
        rest=timeString[1].strip()
        #hours=int(days)*24
        lastruntime= dt.datetime.strptime(rest,"%H:%M:%S")
        #one day less
        #lastruntime=lastruntime+dt.timedelta(days=days-1)
        deltaTime = dt.timedelta(days=days,hours=lastruntime.hour, minutes=lastruntime.minute, seconds=lastruntime.second)
    
    endTime=lastST+deltaTime
    #remove one extra day
    #endTime=dt.timedelta(days=endTime.day-1,hours=endTime.hour, minutes=endTime.minute, seconds=endTime.second)
    
            
    #generate summary
    summary='\n<h2> Summary </h2>'
    summary+='\n<div class="summary">'
    summary+='\n<code>Time start: {}     Time end: {}      Total time: {}</code>'.format(str(startTime),str(endTime), str(endTime-startTime))
    summary+='\n<br><br>'
    summary+='\n<span>Num commands: {}</span>'.format(numCommands)
    summary+='\n<br><br>'
    summary+='\nNum failed commands: {}'.format(failedCommands)
    summary+='\n<br><br>'
    summary+='\nNum passed commands: {}'.format(passedCommands)
    summary+='\n<br><br>'
    summary+='\nTotal programs: {}'.format(numPrograms)
    summary+='\n<br><br>'
    summary+='\nPrograms: {}'.format(",".join(progNames))
    summary+='\n<br><br>'
    summary+='\n</div>'
        
    envTable=generateEnvReportTable(sysInfo,progList)
    headHTML=headHTML+summary+fullHTML+envTable+"\n</body>\n</html>"
    return headHTML
    

def writeHtmlToPdf(htmlText,outFile):
    if not outFile.endswith(".pdf"):
        outFile=outFile+".pdf"
   
    #read css
    cssFile = pkg_resources.read_text(report_templates, 'simple.css')
    #HTML(string=htmlText).write_pdf(outFile)
    #HTML(string=htmlText).write_pdf(outFile, stylesheets=[CSS('/home/usingh/work/urmi/hoap/pyrpipe/pyrpipe/report_templates/simple.css')])
    HTML(string=htmlText).write_pdf(outFile, stylesheets=[CSS(string=cssFile)])
    print("Report written to {}".format(outFile))
    
def writeHtml(htmlText,outFile):
    if not outFile.endswith(".html"):
        outFile=outFile+".html"
    #read css
    #cssFile = pkg_resources.read_text(report_templates, 'simple.css')
    #HTML(string=htmlText).write_pdf(outFile)
    #HTML(string=htmlText).write_pdf(outFile, stylesheets=[CSS('/home/usingh/work/urmi/hoap/pyrpipe/pyrpipe/report_templates/simple.css')])
    #HTML(string=htmlText).write_pdf(outFile, stylesheets=[CSS(string=cssFile)])
    f=open(outFile,'w')
    f.write(htmlText)
    f.close()
    
    print("Report written to {}".format(outFile))
    

def writeHtmlToMarkdown(htmlText,outFile):
    pass
    """if not outFile.endswith(".md"):
        outFile=outFile+".md"
    
    f=open(outFile,'w')
    f.write(mdText)
    f.close()
    
    print("Report written to {}".format(outFile))
    """
    

def getCommandsFromLog(inFile,filterList,coverage):
    with open(inFile) as f:
        data=f.read().splitlines()
    commands=[]
    for l in data:
        if not l.startswith("#"):
            thisLog=json.loads(l)
            thisName=thisLog["cmd"].split(' ')[0]
            if filterList and thisName in filterList:
                continue
            status=int(thisLog['exitcode'])
            #skip passed
            if coverage=='i' and status==0:
                continue
            #skip failed
            if coverage=='p' and status!=0:
                continue
            
            commands.append(thisLog["cmd"])
            
    return commands

def getStdoutFromLog(inFile,filterList,coverage):
    """Return a dict with objid_program as key and stdout
    """
    with open(inFile) as f:
        data=f.read().splitlines()
    stdout={}
    duplicate_ctr={}
    for l in data:
        if not l.startswith("#"):
            thisLog=json.loads(l)
            thisObj=thisLog['objectid']
            thisProgram=thisLog['commandname']
            #filter program
            thisName=thisLog["stdout"].split(' ')[0]
            if filterList and thisName in filterList:
                continue
            status=int(thisLog['exitcode'])
            #skip passed
            if coverage=='i' and status==0:
                continue
            #skip failed
            if coverage=='p' and status!=0:
                continue
            
            key=thisObj+"_"+thisProgram
            #handle duplicate
            if key in stdout:
                
                if key in duplicate_ctr:
                    duplicate_ctr[key]+=1
                else:
                    duplicate_ctr[key]=1
                #new key    
                suffix=duplicate_ctr[key]
                key=key+"_"+str(suffix)
                
            stdout[key]=thisLog["stdout"]
            
    return stdout



def generateBashScript(logFile,outFile,filterList,coverage='a'):
    commands=getCommandsFromLog(logFile,filterList,coverage)
    if not outFile.endswith(".sh"):
        outFile=outFile+".sh"
    shebang="#!/bin/bash "
    #write to file
    f=open(outFile,"w")
    f.write(shebang+"\n")
    f.write("\n".join(commands))
    
    print("shell commands written to "+outFile)


def checkEnvLog(logFile):
    #check all logs exist
    logFileDir=pu.get_file_directory(logFile)
    basename=pu.get_file_basename(logFile)
    envLog=os.path.join(logFileDir,basename+"ENV.log")
    if not pu.check_files_exist(logFile,envLog):
        print("Please check missing log files. Exiting.")
        sys.exit(1)
    return envLog

def generateMultiqcReport(logFile,filterList,tempDir,outDir="",coverage='a',verbose=False,cleanup=False):
    #dump stdout from logs to temp directory
    stdout=getStdoutFromLog(logFile,filterList,coverage)
    
    flist=[]
    for o in stdout:
        thisName=o+".txt"
        tempFile=os.path.join(tempDir,thisName)
#        print("opening:"+tempFile)
        f=open(tempFile,"w")
        f.write(stdout[o])
        #rint(stdout[o])
        f.close()
        flist.append(tempFile)
    
    #run multiqc
    #tempDir stores .txt files for MQC to read
    mc.run(analysis_dir=tempDir,outdir=outDir)
    
    #cleanup
    if cleanup:
        for f in flist:
            pu.print_blue("Removing {}".format(f))
            os.remove(f)   

def generateBenchmarkReport(logFile,envLog,filterList,tempDir,outFile="",verbose=False):
    """
    ignores failed commands with exitcode !=0
    """
    
    ob=bm.Benchmark(logFile,envLog,out_dir=tempDir)
    #generate benchmarks
    ob.plot_time_perobject()
    ob.plot_time_perprogram()
    
    pu.print_green("Benchmark report saved to:"+tempDir+"/benchmark_reports")


def report():
    
    parser = argparse.ArgumentParser(
   
            description='pyrpipe diagnostic utility\nGenerate analysis report.',
            
            usage='''pyrpipe_diagnostic report [<args>] <logfile>
                    
                    ''')    
    parser.add_argument('-o', help='out file \ndefault: same as input logfile',action="store")
    parser.add_argument('-e', help='report output type: [md,pdf,html] \ndefault: pdf',default='pdf',action="store")
    parser.add_argument('-c',help='Report options [(f)ull,fa(i)l,(p)ass]\ndefault: f',default='f',action="store")
    parser.add_argument('-v',help='verbose',action="store_true")
    parser.add_argument('logfile', help='The log file generated by pyrpipe',action="store")
    args = parser.parse_args(sys.argv[2:])
    
    logFile=args.logfile
    envLog=checkEnvLog(logFile)    
    #parse args
    vFlag=args.v
    if vFlag:
        print("Generating report")
    outFile=""
    if args.o is None:
        outFile=pu.get_file_basename(args.logfile)
    else:
        outFile=args.o
    outFile+='.'+args.e
    
    
    if args.e in ['pdf','html','md']:
        htmlReport=generateHTMLReport('simpleDiv.html',logFile,envLog,coverage=args.c)
        if args.e=='pdf':
            writeHtmlToPdf(htmlReport,outFile)
        elif args.e=='html':
            writeHtml(htmlReport,outFile)
        elif args.e == 'md':
            writeHtmlToMarkdown(htmlReport,outFile)
    else:
        pu.print_boldred("unknown extension:"+args.e+". Exiting")
    
    
    
def shell():
    print("Generating bash script")
    parser = argparse.ArgumentParser(
   
            description='pyrpipe diagnostic utility\nGenerate shell script.',
            
            usage='''pyrpipe_diagnostic report [<args>] <logfile>
                    
                    ''')    
    parser.add_argument('-o', help='out file \ndefault: same as input logfile',action="store")
    parser.add_argument('-c',help='Dump command options [(a)ll,fa(i)l,(p)ass]\ndefault: a',default='a',action="store")
    parser.add_argument('-v',help='verbose',action="store_true")
    parser.add_argument('-f',help='Filter by programs. Provide a comma-separated list e.g., prefetch,STAR,bowtie2 \ndefault None')
    parser.add_argument('logfile', help='The log file generated by pyrpipe',action="store")
    args = parser.parse_args(sys.argv[2:])
    
    logFile=args.logfile  
    #parse args
    vFlag=args.v
    if vFlag:
        print("Generating report")
    outFile=""
    if args.o is None:
        outFile=pu.get_file_basename(logFile)
    else:
        outFile=args.o
    outFile+='.sh'
    
    filters=[]
    if args.f is not None:
        filters= args.f.split(',')
    
    generateBashScript(logFile,outFile,filters,args.c)
    

def benchmark():
    print("Generating benchmarks")
    parser = argparse.ArgumentParser(
   
            description='pyrpipe diagnostic utility\nGenerate benchmark report.',
            
            usage='''pyrpipe_diagnostic report [<args>] <logfile>
                    
                    ''')    
    parser.add_argument('-o', help='out file \ndefault: same as input logfile',action="store")
    parser.add_argument('-e', help='report output type: [MD,PDF,HTML] \ndefault: PDF',default='PDF',action="store")
    parser.add_argument('-v',help='verbose',action="store_true")
    parser.add_argument('-f',help='Filter by programs. Provide a comma-separated list e.g., prefetch,STAR,bowtie2 \ndefault None')
    parser.add_argument('-t',help='Temporary directory. \ndefault ./tmp',action="store")
    parser.add_argument('logfile', help='The log file generated by pyrpipe',action="store")
    args = parser.parse_args(sys.argv[2:])
    
    logFile=args.logfile
    envLog=checkEnvLog(logFile)    
    #parse args
    vFlag=args.v
    if vFlag:
        print("Generating benchmarks")
    outFile=""
    if args.o is None:
        outFile=pu.get_file_basename(args.logfile)
    else:
        outFile=args.o
    outFile+='.'+args.e
    
    filters=[]
    if args.f is not None:
        filters= args.f.split(',')
    #create temp dir
    tempDir=""
    if args.t is not None:
        tempDir= args.t
    else:
        tempDir=os.path.join(os.getcwd(),"tmp")
    #create tmp dir
    if not pu.check_paths_exist(tempDir):
        pu.mkdir(tempDir)
        
    generateBenchmarkReport(logFile,envLog,filters,tempDir,outFile=outFile,verbose=args.v)

    
def multiqc():
    print("Generating html report with multiqc")
    parser = argparse.ArgumentParser(
   
            description='pyrpipe diagnostic utility\nGenerate report with multiqc.',
            
            usage='''pyrpipe_diagnostic multiqc [<args>] <logfile>
                    
                    ''')    
    parser.add_argument('-o', help='out directory \ndefault: <./>',action="store")
    parser.add_argument('-c',help='Dump command options [(a)ll,fa(i)l,(p)ass]\ndefault: a',default='a',action="store")
    parser.add_argument('-v',help='verbose',action="store_true")
    parser.add_argument('-f',help='Filter by programs. Provide a comma-separated list e.g., prefetch,STAR,bowtie2 \ndefault None')
    parser.add_argument('-t',help='Temporary directory. \ndefault ./tmp',action="store")
    parser.add_argument('-r',help='Remove stdout files after processing. \ndefault ./tmp',action="store_true")
    parser.add_argument('logfile', help='The log file generated by pyrpipe',action="store")
    args = parser.parse_args(sys.argv[2:])
    
    logFile=args.logfile
    
    #parse args
    vFlag=args.v
    if vFlag:
        print("Generating MutiQC report")
    outDir=""
    if args.o is None:
        outDir=os.getcwd()
    else:
        outDir=args.o
    
    
    filters=[]
    if args.f is not None:
        filters= args.f.split(',')
    
    #create temp dir
    tempDir=""
    if args.t is not None:
        tempDir= args.t
    else:
        tempDir=os.path.join(os.getcwd(),"tmp")
    #create tmp dir
    if not pu.check_paths_exist(tempDir):
        pu.mkdir(tempDir) 
    
    #run multiqc
    generateMultiqcReport(logFile,filters,tempDir,outDir=outDir,coverage=args.c,verbose=args.v,cleanup=args.r)


##Start parsing
subcommands=['report','shell','benchmark','multiqc','all']
parser = argparse.ArgumentParser(
            
            description='pyrpipe diagnostic utility',
            
            usage='''pyrpipe_diagnostic <command> [<args>] <logfile>
                    The commands are:
                    report     Generate analysis report
                    shell      Generate all commands to shell (bash) script
                    benchmark Generate bemchmarks
                    multiqc Generate HTML report using multiqc
                    
                    ''')
parser.add_argument('command', help='Subcommand to run [report,bash,benchmark,all]')


#parse first and last argument as subcommand and logfile 
args = parser.parse_args(sys.argv[1:2])
if args.command not in subcommands:
    print ('Unrecognized command')
    parser.print_help()
    exit(1)

if args.command == 'report':
    report()
elif args.command == 'shell':
    shell()
elif args.command == 'benchmark':
    benchmark()
elif args.command == 'multiqc':
    multiqc()


    
    
    
    
    
    
    
    
    
    
    
    
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 21 12:04:28 2019

@author: usingh
"""

import os
import datetime as dt





class Colors:
    """
    Colors used for printing messages.
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    LightMagenta = "\033[95m"
    LightYellow  = "\033[93m"
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def print_boldred(text):
    """Print in bold red font
    
    Parameters
    ----------
    
    text: str
        text to print
        
    Returns:   None
    """
    print (Colors.FAIL + Colors.BOLD+ text + Colors.ENDC)

def print_green(text):
    """Print in green font
    
    Parameters
    ----------
    
    text: str
        text to print
        
    Returns:   None
    """
    print (Colors.OKGREEN + text + Colors.ENDC)

def print_blue(text):
    """Print in blue font
    
    Parameters
    ----------
    
    text: str
        text to print
        
    Returns:   None
    """
    print (Colors.OKBLUE + text + Colors.ENDC) 

def print_magenta(text):
    """Print in magenta font
    
    Parameters
    ----------
    
    text(str): text to print
    
    Returns:   None
    """
    print (Colors.LightMagenta + text + Colors.ENDC) 
    
def print_info(text):
    """Print an info message
    
    Parameters
    ----------
    
    text: str
        text to print

    Returns:   None
    """
    print_magenta(text)
    
def print_yellow(text):
    """Print in yellow font
    
    Parameters
    ----------
    
    text: str
        text to print
        
    Returns:   None
    """
    print (Colors.LightYellow + text + Colors.ENDC) 

######End color functions###################

def get_time_stamp(shorten=False):
    """Function to return current timestamp.
    
    Parameters
    ----------
    
    shorten: bool
        return short version without space, dash and colons

    :return: timestamp as string
    :rtype: string
    """
    
    timestamp=str(dt.datetime.now()).split(".")[0].replace(" ","-")
    if shorten:
        timestamp=timestamp.replace("-","").replace(" ","").replace(":","")
    return timestamp
    

def get_sra_ftppath(srrid):
    """Return an ftp address to download sra files
    """
    if len(srrid) <6:
        return None
    
    parent_path="anonftp@ftp.ncbi.nlm.nih.gov:/sra/sra-instant/reads/ByRun/sra/"
    parent_path=parent_path+srrid[0:3]+"/"+srrid[0:6]+"/"+srrid+"/"+srrid+".sra"
    
    return parent_path
    


def check_paths_exist(*args):
    """Function to check if a directory exists.
    
    Parametrs
    ---------
    
    args: tuple
        a list of paths to check

    :return: return true only if all paths exist
    :rtype: bool
    """
    fail_flag=False
    for path in args:
        if not (os.path.exists(path) and os.path.isdir(path)):
            fail_flag=True
    if fail_flag==True:
        return False
    return True


def check_files_exist(*args):
    """Function to check if files exist.
    
    Parametrs
    ---------
    
    args: tuple
        a list of paths to check

    :return: return true only if all files exist
    :rtype: bool
    """
    fail_flag=False
    for path in args:
        if not os.path.isfile(path):
            #print_boldred("File not found: "+path)
            fail_flag=True
    
    if fail_flag:
        return False
    return True

def check_hisatindex(index):
    """Function to check if hisat2 index is valid and exists.
    
    Parameters
    ----------
    
    index: str
        Path to the index 

    :return: Return true if index is valid
    :rtype: bool
    """
    return check_files_exist(index+".1.ht2")

def check_salmonindex(index):
    """Function to check if salmon index is valid and exists.
    
    Parameters
    ----------
    
    index: str
        Path to the index 

    :return: Return true if index is valid
    :rtype: bool
    """
    if not check_paths_exist(index):
        return False
    return True

def check_starindex(index):
    """Function to check if star index is valid and exists.
    
    Parameters
    ----------
    
    index: str
        Path to the index 

    :return: Return true if index is valid
    :rtype: bool
    """
    if check_paths_exist(index):
        files_to_check=['chrLength.txt',
                      'chrNameLength.txt',
                      'chrName.txt',
                      'chrStart.txt',
                      'genomeParameters.txt',
                      'Genome']
        for f in files_to_check:
            if not check_files_exist(os.path.join(index,f)):
                return False
        return True
    
    return False

def check_bowtie2index(index):
    """Function to check if bowtie2 index is valid and exists.
    
    Parameters
    ----------
    
    index: str
        Path to the index 

    :return: Return true if index is valid
    :rtype: bool
    """
    return check_files_exist(index+".1.bt2")
    

def byte_to_readable(size_bytes):
    """Function to convert bytes to human readable format (MB,GB ...)
    
    Parameters
    ----------
    
    size_bytes: float
        size in bytes

    :return: Return size in human readable format
    :rtype: str
    """
    
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if size_bytes < 1024.0:
            return "%3.1f %s" % (size_bytes, x)
        size_bytes /= 1024.0


def get_file_size(file_path):
    """Returns file size in human readable format
    
    Parameters
    ----------
    
    file_path: str
        Path to file

    :return: Return size in human readable format
    :rtype: string
    """
    
    if (check_files_exist(file_path)):
        file_info = os.stat(file_path)
        return byte_to_readable(file_info.st_size)
    

#TODO: override in case of empty list
def parse_java_args(valid_args_list,passed_args):
    """
    Function creates arguments to pass to java programs
    
    Parameters
    ----------
    
    valid_args_list: list
        list of valid arguments. Invalid arguments will be ignored
    passed_args: *dict
        keyword value argument list to be parsed

    :return: a list with command line arguments to be used with subprocess.popen
    :rtype: list
    
    Examples
    --------
    >>> parse_java_args(['A','B','-C'], {"A": "3", "B": "22","-C":""})
        ['A=3', 'B=22', '-C']
    """
    popen_args=[]
    special_args=["--"]
    positional_args=[]
    
    #empty list supplied consider all armunets valid
    if len(valid_args_list)<1:
        valid_args_list=passed_args.keys()
        #above command will also add specialArgs, remove those
        for x in special_args:
            if x in valid_args_list:
                valid_args_list.remove(x)
    
    for key, value in passed_args.items():
        #check if key is a valid argument
        if key in valid_args_list:
            #do not add emty parameters e.g. -q or -v
            if len(value)>0:
                popen_args.append(key+"="+value)
            else:
                popen_args.append(key)
        elif key in special_args:
            positional_args.extend(value)
        
        else:
            print("Unknown argument {0} {1}. ignoring...".format(key, value))
    popen_args.extend(positional_args)
    return popen_args
    
    
def parse_unix_args(valid_args_list,passed_args):
    """Function creates command line arguments to pass to unix programs
    
    Parameters
    ----------
    
    valid_args_list: list
        list of valid arguments. Invalid arguments will be ignored
    passed_args: *dict
        keyword value argument list to be parsed
        
    :return: a list with command line arguments to be used with subprocess.popen
    :rtype: list
    
    Examples
    --------
    >>> parse_unix_args(['-O','-t','-q'], {"-O": "./test", "Attr2": "XX","--":("IN1","IN2")})
        Unknown argument Attr2 XX. ignoring...
        ['-O', './test', 'IN1', 'IN2']
    """
    
        
    
    popen_args=[]
    """
    Define some special arguments.
    -- to pass input which don't follow any flag e.g. mypro.sh input1 input2
    """
    special_args=["--"]
    positional_args=[]
    
    #empty list supplied consider all armunets valid
    if len(valid_args_list)<1:
        valid_args_list=list(passed_args.keys())
        #above command will also add specialArgs, remove those
        for x in special_args:
            if x in valid_args_list:
                valid_args_list.remove(x)
        
        
    for key, value in passed_args.items():
        #check if key is a valid argument
        if key in valid_args_list:
            popen_args.append(key)
            #do not add emty parameters e.g. -q or -v
            if len(value)>0:
                #if there are multiple values for a flag separated by space
                popen_args.extend(value.split(" "))
                        
                #popen_args.append(value)
        elif key in special_args:
            positional_args.extend(value)
        else:
            print("Unknown argument {0} {1}. ignoring...".format(key, value))
    popen_args.extend(positional_args)
    return popen_args
    




def get_file_directory(file_path):
    """Returns directory of a file
    
    Parameters
    ----------
    
    file_path: str
        Path to file

    :return: directory os file_path
    :rtype: string
    """
    return os.path.split(file_path)[0]

def get_filename(file_path):
    """Returns filename with extension
    
    Parameters
    ----------
    
    file_path: str
        Path to file

    :return: filename
    :rtyep: string
    """
    return os.path.split(file_path)[1]

def get_fileext(file_path):
    """Returns file extension
    
    Parameters
    ----------
    
    file_path: str
        Path to file

    :return: file extension
    :rtype: string
    """
    return os.path.splitext(file_path)[1]

def get_file_basename(file_path):
    """Returns file basename without extension
    
    Parameters
    ----------
    
    file_path: str
        Path to file

    :return: file basename without extension
    :rtype: string
    """
    return os.path.splitext(get_filename(file_path))[0]
    



def mkdir(dir_path):
    """Create a directory

    :return: true is directory created
    :rtype: bool
    """
    #print("creating Dir:"+dir_path)
    try:
        #os.mkdir(dir_path)
        os.makedirs(dir_path)
    except OSError:
        return False
    return True


def get_union(*args):
    """Return unioin of multiple input lists.
    """
    return list(set().union(*args)) 
    
#def scan_SRA_dir(path):
#    pass


